/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <gtk/gtkwidget.h>
#include <gtk/gtkstock.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtknotebook.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkhbbox.h>

#include "intl.h"
#include "logging.h"

#include "utils.h"
#include "ui.h"
#include "gtk-utils.h"

#include "gtk-about.h"

/* About panel */
static void
GtkAbout_create(XePUIMain *uimain);

static void
GtkAbout_ok_cb(GtkButton *button, gpointer data);

static gboolean
GtkAbout_delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data);

static GtkWidget *
GtkAbout_append_about_page(GtkWidget *notebook);

static GtkWidget *
GtkAbout_append_text_page(GtkWidget *notebook, 
			  const gchar *title, 
			  const gchar *text);

static void 
GtkAbout_browse_homepage_cb(GtkButton *button, gpointer data);



/* About Panel */
void
a_GtkAbout_show(XePUIMain *uimain)
{
  if (!uimain->about)
    GtkAbout_create(uimain);
  else
    gtk_window_present(GTK_WINDOW(uimain->about));
}

static void
GtkAbout_create(XePUIMain *uimain)
{
  GtkWidget *aboutw;
  GtkWidget *mainvbox;
  GtkWidget *image;
  GtkWidget *notebook;
  GtkWidget *hbuttonbox;
  GtkWidget *ok_btn;
  GtkWidget *label;
  gchar *name;

  A_LOGGING_TRACE;
  
  /* Main window */
  aboutw = a_GtkUtils_transient_new(uimain->toplevel,
				    _("About XetPan"),
				    400, 400);

  /* Main Vertical Box */
  mainvbox = gtk_vbox_new(FALSE, 12);
  gtk_container_add(GTK_CONTAINER(aboutw), mainvbox);

  /* Icon */
  image = gtk_image_new_from_file(PKGDATADIR "/pixmaps/xetpan.png");
  gtk_box_pack_start(GTK_BOX(mainvbox), image, FALSE, FALSE, 0);
  
  /* Program name and version */
  name = g_strdup_printf(_("XetPan %s"), VERSION);
  label = a_GtkUtils_label_new_bold_large(name);
  g_free(name);

  gtk_box_pack_start(GTK_BOX(mainvbox), label, FALSE, FALSE, 0);

  /* Notebook */
  notebook = gtk_notebook_new();
  gtk_box_pack_start(GTK_BOX(mainvbox), notebook, TRUE, TRUE, 0);
  
  /* Button Box */
  hbuttonbox = gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(hbuttonbox), GTK_BUTTONBOX_END);
  gtk_box_pack_start(GTK_BOX(mainvbox), hbuttonbox, FALSE, FALSE, 0);

  /* OK button */
  ok_btn = gtk_button_new_from_stock(GTK_STOCK_OK);
  gtk_container_add(GTK_CONTAINER(hbuttonbox), ok_btn);

  /* About page */
  GtkAbout_append_about_page(notebook);

  /* Authors page */
  GtkAbout_append_text_page(notebook, _("Authors"), A_UI_ABOUT_AUTHORS);

  /* License page */
  GtkAbout_append_text_page(notebook, _("License"), A_UI_ABOUT_LICENSE);

  /* Signals */
  g_signal_connect(G_OBJECT(ok_btn), "clicked", 
                   G_CALLBACK(GtkAbout_ok_cb),
                   aboutw);
  g_signal_connect(G_OBJECT(aboutw), "delete-event",
                   G_CALLBACK(GtkAbout_delete_cb),
                   aboutw);

  uimain->about = (void *) aboutw;

  gtk_widget_show_all(aboutw);
}

static void
GtkAbout_ok_cb(GtkButton *button, gpointer data)
{
  gtk_widget_hide(GTK_WIDGET(data));
}

static gboolean
GtkAbout_delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_widget_hide(GTK_WIDGET(data));
  return TRUE; /* i.e. do not destroy */
}

static GtkWidget *
GtkAbout_append_about_page(GtkWidget *notebook)
{
  GtkWidget *vbox;
  GtkWidget *label;
  GtkWidget *homepage_btn;

  label = gtk_label_new(_("About"));

  vbox = gtk_vbox_new(FALSE, 12);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, label);

  /* Short description */
  label = gtk_label_new(_("A mail and news client."));
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

  /* Home page button */
  homepage_btn = a_GtkUtils_button_new_from_stock("gtk-home", 
						  A_UI_ABOUT_HOME_PAGE,
						  TRUE,
						  GTK_ICON_SIZE_BUTTON);

  gtk_box_pack_start(GTK_BOX(vbox), homepage_btn, FALSE, FALSE, 0);
  gtk_button_set_relief(GTK_BUTTON(homepage_btn), GTK_RELIEF_NONE);

  /* Copyright */
  label = gtk_label_new(gettext(A_UI_ABOUT_COPYRIGHT));
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

  /* Signals */
  g_signal_connect(G_OBJECT(homepage_btn), "clicked", 
                   G_CALLBACK(GtkAbout_browse_homepage_cb), NULL);
  return vbox;
  
}

static void 
GtkAbout_browse_homepage_cb(GtkButton *button, gpointer data)
{
  gchar *cmd;
  cmd = g_strdup_printf("dillo %s", A_UI_ABOUT_HOME_PAGE);
  a_Utils_exec_cmd(cmd, 0, 0, 0, NULL, NULL, 0);
  g_free(cmd);
}

static GtkWidget *
GtkAbout_append_text_page(GtkWidget *notebook, 
			  const gchar *title, 
			  const gchar *text)
{
  GtkWidget *label;
  GtkWidget *scrollw;
  GtkWidget *text_view;

  label = gtk_label_new(title);

  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_container_set_border_width(GTK_CONTAINER(scrollw), 12);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw), 
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollw), 
                                      GTK_SHADOW_ETCHED_IN);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), scrollw, label);

  text_view = a_GtkUtils_text_view_new(gettext(text), -1, FALSE);

  gtk_container_add(GTK_CONTAINER(scrollw), text_view);

  return scrollw;
}

#endif
