/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef GTK_COMPOSEWINDOW_H
#define GTK_COMPOSEWINDOW_H

#include <gtk/gtkwidget.h>

#include "xetpan-types.h" /* XetPan */
#include "message-types.h" /* XePMessage */
#include "ui-types.h"     /* XePUIType, XePUIMain */
#include "gtk-types.h"
#include "mta.h"

/* Defined here because ui.h needs it to create an union */
struct XePGTKComposeWindow_s
{
  /*! The following members must be the same as the ones in
   *  XePUIMessageViewAny defined in ui.h
   */
  XePUIType type;                  /**< UI type */
  XetPan *xetpan;                  /**< The global XetPan data structure */
  XePUIMain *uimain;               /**< The parent main window, if any */
  XePMTA *mta;                     /**< The selected MTA */
  XePGTKMessageView *messageview;  /**< The related message view */
  char *sent_folder;               /**< The folder that stores outgoing msg */
  /*! Private to the Gtk+ UI */
  GtkWidget *toplevel; /**< Used when the viewer is a top-level one */
  GtkWidget *menubar;
  XePGTKToolbar *toolbar;
  GtkWidget *header_table;
  GtkWidget *from_combo;
  GtkWidget *subject_entry;
  GList *headers;
  int header_pane_position;
};

XePGTKComposeWindow *
a_GtkComposewindow_create(XetPan *xetpan, char *terminal);

void
a_GtkComposewindow_free(XePGTKComposeWindow *uicompose);

void
a_GtkComposewindow_add_header(XePGTKComposeWindow *uicompose, 
			      XePMessageHeaderType header_type,
			      const char *header_name,
			      const char *header_value);

void
a_GtkComposewindow_add_to_header(XePGTKComposeWindow *uicompose,
				 const char *value);

void
a_GtkComposewindow_add_cc_header(XePGTKComposeWindow *uicompose,
				 const char *value);

void
a_GtkComposewindow_add_bcc_header(XePGTKComposeWindow *uicompose,
				  const char *value);

void
a_GtkComposewindow_add_other_header(XePGTKComposeWindow *uicompose,
				    const char *other_header,
				    const char *value);

void
a_GtkComposewindow_set_message(XePGTKComposeWindow *uicompose, 
			       XePMessage *message);
/*! Set the initial body text of the to be edited message 
 *  \param uicompose the concerned XePGTKComposeWindow
 *  \param text the initial text
 *  \param text_size its size
 *  \param text_charset its character set
 */
void
a_GtkComposewindow_set_body_text(XePGTKComposeWindow *uicompose,
				 const char *text,
				 size_t text_size,
				 const char *text_charset);

/*! Populate the header entries of the compose window with
 *  the initial message
 *  \param uicompose the concerned XePGTKComposeWindow
 *  \param message the XePMessage of which the headers
 *         are to be shown
 */
void
a_GtkComposewindow_list_headers(XePGTKComposeWindow *uicompose);

/*! Build and send the composed message
 *  \param uicompose the concerned XePGTKComposeWindow
 */
void
a_GtkComposewindow_send(XePGTKComposeWindow *uicompose);

#endif
