/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/poll.h>
#include <unistd.h>

#include <gtk/gtk.h>

#include <gtk/gtkalignment.h>

#include <gdk/gdkkeysyms.h>
#include <gdk/gdkevents.h> /* GDK_BUTTON_PRESS */

#include <clist.h>

#include "intl.h"

#include "x_alloc.h"
#include "defines.h"
#include "logging.h"
#include "remote.h"
#include "welcome_dialog.h"
#include "pixmaps/xetpan.xpm"
#include "foldersview.h"
#include "ui.h"
#include "gtk-ui.h"
#include "gtk-utils.h"
#include "gtk-log.h"
#include "gtk-about.h"
#include "gtk-messageview.h"
#include "gtk-toolbar.h"
#include "gtk-id-editor.h"
#include "manager.h"
#include "identity.h"
#include "conf.h"
#include "utils.h"
#include "gtkxtreeview.h"

#ifdef HAVE_PYTHON_H
#include "gtk-python.h"
#endif

#define XETPAN_TIMES

#ifdef XETPAN_TIMES
#include <sys/times.h>
#endif

#define VIEW_TREE

enum
  {
    FV_FOLDER_COL,
    FV_NEW_COL,
    FV_UNREAD_COL,
    FV_TOTAL_COL, 
    FV_WEIGHT_COL,
    FV_FOLDERSVIEWNODE_COL,
    FV_N_COLUMNS
  };

enum
  {
    SV_NONE,
    SV_STATUS_COL,
    SV_FLAG_COL,
    SV_SUBJECT_COL,
    SV_FROM_COL,
    SV_DATE_COL,
    SV_SIZE_COL,
    SV_TO_COL,
    SV_INDEX_COL
  };

/* TODO: check if XePGTKUI will be needed */
typedef struct XePGTKUI_s XePGTKUI;
struct XePGTKUI_s
{
};

static void
GtkUI_add_stock_icons(void);

static void
GtkUI_icon_factory_add_icon(GtkIconFactory *icon_factory,
			    gchar *icon_name,
			    gchar *stock_id);

/* Main View: menu bar, toolbar, folderview, summaryview, messageview */

static GtkWidget *
GtkUI_mainview_menu_create(XePUIMain *uimain);

static void
GtkUI_mainview_free(XePUIMain *uimain);

static void
GtkUI_messageview_popup_cb(GtkTextView *textview, 
			   GtkMenu *menu, 
			   gpointer data);

static void
GtkUI_mainview_file_new_cb(gpointer data,
			   guint action,
			   GtkWidget *widget);

static void
GtkUI_mainview_file_new_real_cb(void *axetpan,
				int response,
				gchar *display);

static void
GtkUI_mainview_file_close_cb(gpointer data,
			     guint action,
			     GtkWidget *widget);

static void
GtkUI_mainview_edit_identities_cb(gpointer data,
				  guint action,
				  GtkWidget *widget);
#ifdef HAVE_PYTHON_H
static void
GtkUI_mainview_tools_python_cb(gpointer data,
			       guint action,
			       GtkWidget *widget);
#endif

static void
GtkUI_mainview_tools_log_cb(gpointer data,
			    guint action,
			    GtkWidget *widget);

static void
GtkUI_mainview_about_cb(gpointer data,
		     guint action,
		     GtkWidget *widget);

static void
GtkUI_messageview_fetch_source(XePGTKMessageView *mview, 
			       XePManagerMsgCb callback);

static void
GtkUI_messageview_popup_show_source_cb(GtkMenuItem *menu_item,
				       gpointer data);

static void
GtkUI_messageview_popup_save_source_as_cb(GtkMenuItem *menu_item,
					  gpointer data);

static void
GtkUI_messageview_show_source_cb(void *amview, void *amessage);

static void
GtkUI_messageview_save_source_cb(void *amview, void *amessage);

static gboolean
GtkUI_mainview_delete_cb(GtkWidget *widget,
			 GdkEvent *event,
			 gpointer data);

/* Folder View */

static XePGTKFoldersView *
GtkUI_foldersview_new(XePUIMain *uimain,
		      XePFoldersViewsList *foldersviewslist);

static void
GtkUI_foldersview_free(XePGTKFoldersView *fview);

static void
GtkUI_foldersview_set_cb(GtkMenuItem *menuitem, gpointer data);

static void
GtkUI_foldersview_view(XePGTKFoldersView *fview,
		       XePFoldersView *foldersview);

static void
GtkUI_foldersview_set_node(GtkTreeStore *store, 
			   GtkTreeIter  *iter,
			   XePFoldersViewNode *fnode);

static void
GtkUI_foldersview_append_node(cntree *tree, 
			      GtkTreeStore *store, 
			      GtkTreeIter *parent);

static gboolean
GtkUI_foldersview_button_release_cb(GtkWidget *widget,
				    GdkEventButton *event,
				    gpointer data);

static void
GtkUI_foldersview_select_foldersviewnode(XePGTKFoldersView *fview,
					 XePFoldersViewNode *fnode);
static gboolean
GtkUI_foldersview_select_node(XePGTKFoldersView *fview, 
			      XePFoldersViewNode *fnode,
			      GtkTreeIter *treeiter);

static gboolean
GtkUI_foldersview_refresh_node(GtkTreeModel *model,
			       GtkTreePath  *path,
			       GtkTreeIter  *iter,
			       gpointer      data);
static void
GtkUI_foldersview_refresh(XePGTKFoldersView *fview);

/* Summary View */

static XePGTKSummaryView *
GtkUI_summaryview_new(XePUIMain *uimain);

static void
GtkUI_summaryview_free(XePGTKSummaryView *sview);

static gboolean
GtkUI_summaryview_msg_data_func(mailmessage *msg, 
				guint col_id,
				void **content,
				gchar **attribute,
				gpointer data);


static gboolean
GtkUI_summaryview_tree_data_func(void *iterator, 
				guint col_id,
				void **content,
				gchar **attribute,
				gpointer data);

static void
GtkUI_summaryview_tree_params_func(void *iterator, 
				  gboolean *has_children,
				  guint *depth,
				  gpointer data);

static gboolean
GtkUI_summaryview_tree_iterator_func(void **iterator, 
				     gboolean next,
				     gboolean ignore_children,
				     guint sort_col_id,
				     GtkSortType sort_order,
				     gpointer data);
static gboolean
GtkUI_summaryview_list_data_func(void *iterator, 
				 guint col_id,
				 void **content,
				 gchar **attribute,
				 gpointer data);

static void
GtkUI_summaryview_list_params_func(void *iterator, 
				  gboolean *has_children,
				  guint *depth,
				  gpointer data);

static gboolean
GtkUI_summaryview_list_iterator_func(void **iterator, 
				     gboolean next,
				     gboolean ignore_children,
				     guint sort_col_id,
				     GtkSortType sort_order,
				     gpointer data);

static mailmessage *
GtkUI_summaryview_get_selected_msg(XePGTKSummaryView *sview);

static gboolean
GtkUI_summaryview_button_release_cb(GtkWidget *widget,
				    GdkEventButton *even,
				    gpointer data);
static void
GtkUI_summaryview_refresh(XePGTKSummaryView *sview);

static void
GtkUI_summaryview_update_selected(XePGTKSummaryView *sview);

static void
GtkUI_summaryview_delete_item(XePGTKSummaryView *sview, mailmessage *msg);

static void GtkUI_summaryview_update_cb(void *asview, void *amessage);

/* libEtPan! extensions */
static struct mail_envelopes_tree *
mail_envelopes_tree_next(struct mail_envelopes_tree *self,
			 struct mail_envelopes_tree *child,
			 int ignore_children);
static struct mail_envelopes_tree *
mail_envelopes_tree_prev(struct mail_envelopes_tree *parent,
			 struct mail_envelopes_tree *self);

static uint32_t
mail_envelopes_list_next(struct mail_envelopes_list *ep_list, uint32_t index);

static uint32_t
mail_envelopes_list_prev(struct mail_envelopes_list *ep_list, uint32_t index);


/* IO Watch Callback */

static gboolean
GtkUI_io_watch_cb(GIOChannel *channel, GIOCondition gcondition, gpointer data);

/******************************************************************************/
/*                                                                            */
/* Basic GUI functions                                                        */
/*                                                                            */
/******************************************************************************/

/* Init display and abort if fail */
void
a_GtkUI_init(XetPan *xetpan, int *argc, char ***argv)
{
  A_LOGGING_TRACE;
  if (a_GtkUI_init_check(xetpan, argc, argv))
    {
      A_LOGGING_ERROR("Unable to initilialize GUI. Aborting");
      exit(1);
    }
}

/* Init display and return -1 if fail */
int
a_GtkUI_init_check(XetPan *xetpan, int *argc, char ***argv)
{
  int ret;

  A_LOGGING_TRACE;

  ret = 0;

  if (gtk_init_check(argc, argv) == FALSE)
    {
      const char *display_name_arg = gdk_get_display_arg_name ();
      A_LOGGING_ERROR("GTK+ User Interface could not be initialized:");
      A_LOGGING_ERROR("cannot open display %s", 
		      display_name_arg ? display_name_arg : " ");
      ret = -1;
    }
  else
    {
      if (xetpan->ui->gtk)
	{
	  A_LOGGING_INFO("GTK+ User Interface already initialized.");
	}
      else
	{
	  xetpan->ui->gtk = x_alloc(1, XePGTKUI);
	  if (xetpan->ui->gtk == NULL)
	    {
	      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
	      ret = -1;
	    }
          else
            {
              gtk_settings_set_long_property(gtk_settings_get_default(),
                                             "gtk-can-change-accels",
                                             1, 
                                             "XetPan");
	      GtkUI_add_stock_icons();
	      A_LOGGING_INFO("Initialized GTK+ User Interface");
            }
	}
    }
  return ret;
}

/* Main loop */
void
a_GtkUI_main(XetPan *xetpan)
{
  int n_watches, i;
  XePUIIOWatcher *watcher;
  XePGTKUI *gtkui = (XePGTKUI *) xetpan->ui->gtk;

  A_LOGGING_TRACE;
  
  gtk_main();
  
  n_watches = carray_count(xetpan->ui->watched_fd);
  for (i = 0; i < n_watches; i++)
    {
      watcher = (XePUIIOWatcher *) carray_get(xetpan->ui->watched_fd, i);
      a_GtkUI_delete_io_watch(xetpan->ui, watcher);
    }

  x_free(gtkui);
  xetpan->ui->gtk = NULL;
  
  A_LOGGING_DEBUG("xetpan->ui->gtk = %p", xetpan->ui->gtk);

  A_LOGGING_DEBUG("end");
}

void
a_GtkUI_main_quit(XetPan *xetpan)
{
  A_LOGGING_TRACE;
  gtk_main_quit();
}

int
a_GtkUI_events_pending(XetPan *xetpan)
{
  return gtk_events_pending();
}

void
a_GtkUI_main_iteration(XetPan *xetpan)
{
  gtk_main_iteration();
}

static void
GtkUI_add_stock_icons(void)
{
  GtkIconFactory *icon_factory;

  icon_factory = gtk_icon_factory_new();
  

  GtkUI_icon_factory_add_icon(icon_factory, "xetpan.png",
			      "xetpan-icon");
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-status.png",
			      "xetpan-status");
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-flag.png",
			      "xetpan-flag");
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-reply.png",
			      "xetpan-reply");
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-forward.png",
			      "xetpan-forward");
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-replied.png",
			      "xetpan-replied");
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-forwarded.png",
			      "xetpan-forwarded");
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-replied-forwarded.png",
			      "xetpan-replied-forwarded");

  gtk_icon_factory_add_default(icon_factory);
}

static void
GtkUI_icon_factory_add_icon(GtkIconFactory *icon_factory,
			    gchar *icon_name,
			    gchar *stock_id)
{
  GtkIconSet *icon_set;
  GtkIconSource *icon_source;
  gchar *filename;

  icon_set = gtk_icon_set_new();

  icon_source = gtk_icon_source_new();

  filename = g_strdup_printf("%s/pixmaps/%s", PKGDATADIR, icon_name);

  gtk_icon_source_set_filename(icon_source, filename);

  g_free(filename);

  gtk_icon_set_add_source(icon_set, icon_source);

  gtk_icon_factory_add(icon_factory, stock_id, icon_set);
  gtk_icon_source_free(icon_source);

}


/******************************************************************************/
/*                                                                            */
/* Windows                                                                    */
/*                                                                            */
/******************************************************************************/

void
a_GtkUI_first_launch(XetPan *xetpan)
{
  A_LOGGING_DEBUG("not implemented yet");
  welcome_dialog(xetpan, 0);
}

void
a_GtkUI_compose_pending_requests(XetPan *xetpan)
{
  A_LOGGING_DEBUG("not implemented yet");
}

/** Create a new Gtk+ main view 
 *  It should not be called directly: it must be called by
 *  a_UI_mainview_create() so the mainview is registered
 */
XePUIMain *
a_GtkUI_mainview_create(XetPan *xetpan, char *terminal)
{
  XePUIMain *uimain;
  XePGTKFoldersView *fview;
  XePGTKSummaryView *sview;
  XePGTKMessageView *mview;
  GtkWidget *toplevel;
  GtkWidget *topvbox;
  GtkWidget *menubar;
#if WITH_TOOLBAR_HANDLE
  GtkWidget *toolbar_handle;
#endif
  GtkWidget *toolbar_hbox;
  GtkWidget *main_frame;
  GtkWidget *fview_frame;
  GtkWidget *message_frame;
  GtkWidget *vertical_pane;
  GtkWidget *horizontal_pane;

  uimain = a_UI_mainview_new();
  if (uimain == NULL)
    return NULL;

  uimain->type = XETPAN_UI_GTK;
  uimain->xetpan = xetpan;

  toplevel = a_GtkUtils_toplevel_new(terminal,
				     _("XetPan"), 
				     xetpan_xpm,
				     620, 460);
  uimain->toplevel = toplevel;

  topvbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(topvbox);
  gtk_container_add(GTK_CONTAINER(toplevel), topvbox);

  /* Menu bar */

  menubar = GtkUI_mainview_menu_create(uimain);
  gtk_widget_show(menubar);
  gtk_box_pack_start(GTK_BOX(topvbox), menubar, FALSE, FALSE, 0);

  uimain->menubar = menubar;

  /* Toolbar */

#ifdef  WITH_TOOLBAR_HANDLE
  toolbar_handle = gtk_handle_box_new();
  gtk_widget_show(toolbar_handle);
  gtk_handle_box_set_shadow_type(GTK_HANDLE_BOX(toolbar_handle), 
				 GTK_SHADOW_OUT);
  gtk_box_pack_start(GTK_BOX(topvbox), toolbar_handle, FALSE, FALSE, 0);
#endif

  toolbar_hbox = a_GtkToolbar_mainview_new(uimain);
  gtk_widget_show(toolbar_hbox);

#ifdef WITH_TOOLBAR_HANDLE
  gtk_container_add(GTK_CONTAINER(toolbar_handle), toolbar_hbox);
#else
  gtk_box_pack_start(GTK_BOX(topvbox), toolbar_hbox, FALSE, FALSE, 0);
#endif

  /* Main Frame */

  main_frame = gtk_frame_new(NULL);
  gtk_widget_show(main_frame);
  gtk_frame_set_shadow_type(GTK_FRAME(main_frame), GTK_SHADOW_OUT);
  gtk_box_pack_start(GTK_BOX(topvbox), main_frame, TRUE, TRUE, 0);
  
  /* Horizontal Pane */

  horizontal_pane = gtk_hpaned_new();
  gtk_widget_show(horizontal_pane);

  /* Vertical Pane */

  vertical_pane = gtk_vpaned_new();
  gtk_widget_show(vertical_pane);

#ifndef NEW_PANES
  gtk_container_add(GTK_CONTAINER(main_frame), horizontal_pane);
  gtk_paned_add2(GTK_PANED(horizontal_pane), vertical_pane);
#else
  gtk_container_add(GTK_CONTAINER(main_frame), vertical_pane);
  gtk_paned_add2(GTK_PANED(vertical_pane), horizontal_pane);
#endif

  /* Folders view */
  /* FIXME: check if fview is NULL (when calloc fails) */

  fview_frame = gtk_alignment_new(0.0, 0.0, 1.0, 1.0);

  gtk_widget_show(fview_frame);

  gtk_paned_add1(GTK_PANED(horizontal_pane), fview_frame);
  
  fview = GtkUI_foldersview_new
    (uimain, xetpan->config->configuration->foldersviewslist);
  uimain->foldersview = fview;

  a_GtkUI_foldersview_show(fview);

  gtk_container_add(GTK_CONTAINER(fview_frame),
		    a_GtkUI_foldersview_widget(fview));

  /* Summary view */

  sview = GtkUI_summaryview_new(uimain);
  uimain->summaryview = sview;

  gtk_paned_pack1(GTK_PANED(vertical_pane),
		  a_GtkUI_summaryview_widget(sview), TRUE, TRUE);

  a_GtkUI_summaryview_show(sview);
  a_GtkUI_foldersview_set_list
    (fview, xetpan->config->configuration->foldersviewslist);

  /* Message view */

  message_frame = gtk_alignment_new(0.0, 0.0, 1.0, 1.0);
  gtk_widget_show(message_frame);

#ifndef NEW_PANES
  gtk_paned_pack2(GTK_PANED(vertical_pane), message_frame, TRUE, TRUE);
#else
  gtk_paned_add2(GTK_PANED(horizontal_pane), message_frame);
#endif

  mview = a_GtkMessageview_create(uimain->xetpan, message_frame, FALSE);
  a_GtkMessageview_link_to_uimain(mview, uimain);

  uimain->messageview = mview;

  /* Python Console */

  uimain->python = NULL;

  /* Log View */

  uimain->logview = a_GtkLog_new(uimain->xetpan, toplevel);

  /* TODO: pane position should be restored from preferences */
  {
    GtkPaned *paned = (GTK_PANED(vertical_pane));
    gtk_paned_set_position(paned, 150);
  }

  /* Signals */

  g_signal_connect(G_OBJECT(toplevel),
		   "delete-event",
		   G_CALLBACK(GtkUI_mainview_delete_cb),
		   uimain);

  g_signal_connect(G_OBJECT(mview->text), "populate-popup",
		   G_CALLBACK(GtkUI_messageview_popup_cb),
		   mview);

  a_GtkUtils_paned_shrink_control(horizontal_pane, &(fview->pane_position));

  a_GtkUtils_paned_shrink_control(vertical_pane, &(sview->pane_position));

  gtk_widget_show(toplevel); 

  a_GtkMessageview_show(mview);

  return uimain;
}

static gboolean
GtkUI_mainview_delete_cb(GtkWidget *widget,
			 GdkEvent *event,
			 gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XetPan *xetpan = (XetPan *) uimain->xetpan;

  A_LOGGING_TRACE;
  if (a_UI_mainview_close(uimain))
    {
      a_XetPan_quit(xetpan);
    }
  return FALSE;
}

void
a_GtkUI_mainview_close(XePUIMain *uimain)
{
  A_LOGGING_TRACE;
  GtkUI_mainview_free(uimain);
}

static void
GtkUI_mainview_free(XePUIMain *uimain)
{
  A_LOGGING_TRACE;
  if (uimain->type != XETPAN_UI_GTK)
    {
      A_LOGGING_SYSTEM_ERROR("Attempt to close a non Gtk+ UI with "
			     "a Gtk+ function");
      return;
    }

  GtkUI_foldersview_free((XePGTKFoldersView *) uimain->foldersview);
  GtkUI_summaryview_free((XePGTKSummaryView *) uimain->summaryview);
  a_UI_messageview_close((XePUIMessageView *) uimain->messageview);
  a_GtkToolbar_mainview_free(uimain);

  /* All the rest is freed when freeing the toplevel widget */
  gtk_widget_destroy((GtkWidget *) uimain->toplevel);
  x_free(uimain);
}

static GtkWidget *
GtkUI_mainview_menu_create(XePUIMain *uimain)
{
  GtkAccelGroup *accel_group;
  GtkItemFactory *item_factory;
  GtkItemFactoryEntry entries[] = {
    { _("/_File"), NULL, NULL, 0, "<Branch>" },
    { _("/File/_New Window"), "<Control>N", GtkUI_mainview_file_new_cb, 0,
      NULL},
    { _("/File/New Window in _Other Display..."), NULL, 
      GtkUI_mainview_file_new_cb, 1, NULL},
    { _("/File/_Close"), "<Control>W", GtkUI_mainview_file_close_cb, 0,
      "<StockItem>", GTK_STOCK_CLOSE},
    { _("/File/_Quit"), "<Control>Q", GtkUI_mainview_file_close_cb, 1,
      "<StockItem>", GTK_STOCK_QUIT},  
    { _("/_Edit"), NULL, NULL, 0, "<Branch>" },
    { _("/Edit/_Identities..."), NULL, GtkUI_mainview_edit_identities_cb, 0,
      NULL},
    { _("/Edit/Prefere_nces..."), NULL, NULL, 0, "<StockItem>", 
      GTK_STOCK_PREFERENCES},
    { _("/_View"), NULL, NULL, 0, "<Branch>" },
    { _("/_Windows"), NULL, NULL, 0, "<Branch>" },
    { _("/_Tools"), NULL, NULL, 0, "<Branch>" },
#if HAVE_PYTHON_H
    { _("/Tools/_Python Console"), NULL, GtkUI_mainview_tools_python_cb, 0},
#endif
    { _("/Tools/_Log Viewer"), NULL, GtkUI_mainview_tools_log_cb, 0, NULL},
    { _("/_Help"), NULL, NULL, 0, "<Branch>" },
    { _("/Help/_Contents"), "F1", NULL, 0, "<StockItem>", GTK_STOCK_HELP},
    { _("/Help/_About"), NULL, GtkUI_mainview_about_cb, 0, NULL}};


  accel_group = gtk_accel_group_new();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<Main>", 
				      accel_group);
  gtk_item_factory_create_items(item_factory, 
				sizeof(entries)/sizeof(entries[0]),
				entries, uimain);
  gtk_window_add_accel_group(GTK_WINDOW(uimain->toplevel), accel_group);
  return gtk_item_factory_get_widget(item_factory, "<Main>");
}

static void
GtkUI_messageview_popup_cb(GtkTextView *textview, GtkMenu *menu, gpointer data)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkWidget *menu_item;

  if (mview->message && mview->message->muid == mview->muid)
    {
      menu_item = gtk_menu_item_new_with_label(_("Show message source"));
      g_signal_connect(G_OBJECT(menu_item), "activate",
		       G_CALLBACK(GtkUI_messageview_popup_show_source_cb),
		       mview);
      gtk_menu_append(menu, menu_item);
      gtk_widget_show(menu_item);

      menu_item = gtk_menu_item_new_with_label(_("Save message source as..."));
      g_signal_connect(G_OBJECT(menu_item), "activate",
		       G_CALLBACK(GtkUI_messageview_popup_save_source_as_cb),
		       mview);
      gtk_menu_append(menu, menu_item);
      gtk_widget_show(menu_item);

    }
}

static void
GtkUI_mainview_file_new_cb(gpointer data,
			   guint action,
			   GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XetPan *xetpan = (XetPan *) uimain->xetpan;
  gchar *display;

  display =  gdk_screen_make_display_name(gtk_window_get_screen
					      (GTK_WINDOW(uimain->toplevel)));
  if (action)
    {
      a_GtkUtils_dialog_get_string(_("Open New Window in Other Display"),
				   uimain->toplevel,
				   _("Display Name:"),
				   NULL,
				   display,
				   XEP_GTKUTILS_OK | XEP_GTKUTILS_CANCEL,
				   0,
				   GtkUI_mainview_file_new_real_cb,
				   xetpan);
    }
  else
    {
      a_UI_mainview_create(xetpan, display, XETPAN_UI_GTK);
    }

  g_free(display);     
}

static void
GtkUI_mainview_file_new_real_cb(void *axetpan,
				int response,
				gchar *display)
{
  XetPan *xetpan = (XetPan *) axetpan;

  if (response == GTK_RESPONSE_OK)
    {
      a_UI_mainview_create(xetpan, display, XETPAN_UI_GTK);
    }
}

static void
GtkUI_mainview_file_close_cb(gpointer data,
			     guint action,
			     GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XetPan *xetpan = (XetPan *) uimain->xetpan;
  if (action || a_UI_mainview_close(uimain))
    a_XetPan_quit(xetpan);
}

static void
GtkUI_mainview_edit_identities_cb(gpointer data,
				  guint action,
				  GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  a_GtkIdEditor_show(uimain);
}

#ifdef HAVE_PYTHON_H
static void
GtkUI_mainview_tools_python_cb(gpointer data,
			       guint action,
			       GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKPython *xeppy = (XePGTKPython *) uimain->python;

  if (xeppy == NULL)
    uimain->python = a_GtkPython_console_new(uimain->xetpan, 
					     GTK_WIDGET(uimain->toplevel));
  else
    a_GtkPython_console_show(xeppy);
}
#endif

static void
GtkUI_mainview_tools_log_cb(gpointer data,
			    guint action,
			    GtkWidget *widget)
{
  a_GtkLog_show((XePGTKLog *)((XePUIMain *) data)->logview);
}

static void
GtkUI_mainview_about_cb(gpointer data,
			guint action,
			GtkWidget *widget)
{
  a_GtkUI_about((XePUIMain *) data);
}

static void
GtkUI_messageview_fetch_source(XePGTKMessageView *mview, 
			       XePManagerMsgCb callback)
{
  XePManagerMsg *msg;
  
  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_FETCH);
  a_Manager_msg_set_source_target(msg, NULL, &mview->message->folder->thread);
  a_Manager_msg_set_args(msg, mview->message, mview->message);
  a_Manager_msg_add_callback(msg, callback, mview);
  a_Manager_msg_submit(msg);
}

static void
GtkUI_messageview_popup_show_source_cb(GtkMenuItem *menu_item, gpointer data)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkUI_messageview_fetch_source(mview, (XePManagerMsgCb) GtkUI_messageview_show_source_cb);
}

static void
GtkUI_messageview_show_source_cb(void *amview, void *amessage)
{
  char *display;
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;
  XePMessage *message = (XePMessage *) amessage;

  display = a_GtkUtils_make_display_name(mview->toplevel);

  a_GtkUtils_file_viewer_new(display, _("Message source"),
			     NULL, 0, 0, NULL, message->data, message->len,
			     message->source_charset);
  x_free(display);
  x_free(message->data);
  message->data = NULL;
  message->len  = 0; 
}

static void
GtkUI_messageview_popup_save_source_as_cb(GtkMenuItem *menu_item,
					  gpointer data)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkUI_messageview_fetch_source(mview, (XePManagerMsgCb) GtkUI_messageview_save_source_cb);
}

static void
GtkUI_messageview_save_source_cb(void *amview, void *amessage)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;
  XePMessage *message = (XePMessage *) amessage;

  a_GtkUtils_file_selection(_("Save message source as..."),
			    mview->toplevel,
			    "email.txt",
			    (XePGtkUtilsSelectionCb) a_Message_data_save_as,
			    message);
			    
}

void
a_GtkUI_mainview_reply(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  XePGTKMessageView *mview = (XePGTKMessageView *) uimain->messageview;
  XePMessage *message;
  gchar *display;
  mailmessage *msg;
  /* Create XePMessage(s) from the selected messages,
   * retrieve their default part, or the one that is
   * currently viewed, then forge a reply
   * This should be done in the ui.c
   */
  msg = GtkUI_summaryview_get_selected_msg(sview);
  if (msg)
    {
      gchar *reply_text;
      const char *reply_charset = NULL;

      message = a_Message_new_for_fetching(sview->folder, msg);
      display = a_GtkUtils_make_display_name(GTK_WIDGET(uimain->toplevel));

      /* TODO: Reply text should be retrieved from the UI 
       * to allow for selective text reply.
       * If the UI does not provide the text to reply to,
       * then use the default MIME part of the message (this
       * is done in ui.c
       */
      /* TODO: Make sure the message viewed is the one to be 
       * replied to.
       */
      reply_text = a_GtkMessageview_get_reply_text(mview);
      if (reply_text )
	{
	  reply_charset = GTK_DEST_CHARSET;
	}

      a_UI_composewindow_reply(uimain->xetpan,
			       display, 
			       uimain->type, 
			       message,
			       reply_text,
			       reply_charset);
      g_free(reply_text);
      g_free(display);
    }
}

void
a_GtkUI_mainview_forward(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  XePMessage *message;
  gchar *display;
  mailmessage *msg;
  /* Create XePMessage(s) from the selected messages,
   * retrieve their default part, or the one that is
   * currently viewed, then forge a reply
   * This should be done in the ui.c
   */
  msg = GtkUI_summaryview_get_selected_msg(sview);
  if (msg)
    {
      message = a_Message_new_for_fetching(sview->folder, msg);
      display = a_GtkUtils_make_display_name(GTK_WIDGET(uimain->toplevel));
      
      a_UI_composewindow_forward(uimain->xetpan, 
				 display, 
				 uimain->type,
				 message);
      g_free(display);
    }
}

void
a_GtkUI_mainview_redirect(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  XePMessage *message;
  gchar *display;
  mailmessage *msg;
  /* Create XePMessage(s) from the selected messages,
   * retrieve their default part, or the one that is
   * currently viewed, then forge a reply
   * This should be done in the ui.c
   */
  msg = GtkUI_summaryview_get_selected_msg(sview);
  if (msg)
    {
      message = a_Message_new_for_fetching(sview->folder, msg);
      display = a_GtkUtils_make_display_name(GTK_WIDGET(uimain->toplevel));
      
      a_UI_composewindow_redirect(uimain->xetpan, 
				  display, 
				  uimain->type,
				  message);
      
      g_free(display);
    }
}

void
a_GtkUI_about(XePUIMain *uimain)
{
  a_GtkAbout_show(uimain);
}

void
a_GtkUI_notify_error(XePUIMain *uimain)
{
  a_GtkToolbar_status_notify_error(uimain);
}

void
a_GtkUI_notify_error_reset(XePUIMain *uimain)
{
  a_GtkToolbar_status_notify_error_reset(uimain);
}

/* Folders View */
XePGTKFoldersView *
GtkUI_foldersview_new(XePUIMain *uimain,
		      XePFoldersViewsList *foldersviewslist)
{
  XePGTKFoldersView *fview;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *label;
  GtkWidget *optmenu;
  GtkWidget *scrollw;
  GtkTreeStore *store;
  GtkWidget *tree;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  fview =  x_alloc(1, XePGTKFoldersView);
  if (fview == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }

  fview->uimain = uimain;

  vbox = gtk_vbox_new(FALSE, 0);

  fview->widget = vbox;

  /* Combox box containing the list of the different views */

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
 
  label = gtk_label_new(_("View: "));
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 8);
  gtk_widget_show(label);

  optmenu = gtk_option_menu_new();
  gtk_box_pack_start(GTK_BOX(hbox), optmenu, TRUE, TRUE, 1);
		   
  gtk_widget_show(optmenu);

  fview->optmenu = optmenu;

  gtk_widget_show(hbox);

  /* Tree widget that contain the virtual tree */

  /* First, the data structure */

  store = gtk_tree_store_new(FV_N_COLUMNS, 
			     G_TYPE_STRING, /**< Folder column       */
			     G_TYPE_STRING, /**< New mail count      */
			     G_TYPE_STRING, /**< Unread mail count   */
			     G_TYPE_STRING, /**< Total mail count    */
			     G_TYPE_INT,    /**< Weight              */
			     G_TYPE_POINTER /**< Pointer to the node */
			     );

  fview->store = store;

  /* The real widget */

  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_box_pack_start(GTK_BOX(vbox), scrollw, TRUE, TRUE, 0);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollw),
				      GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw), 
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);
  gtk_widget_show(scrollw);

  tree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
  gtk_container_add(GTK_CONTAINER(scrollw), tree);
  gtk_tree_view_set_hadjustment(GTK_TREE_VIEW(tree),
				gtk_scrolled_window_get_hadjustment
				(GTK_SCROLLED_WINDOW(scrollw)));
  gtk_tree_view_set_vadjustment(GTK_TREE_VIEW(tree),
				gtk_scrolled_window_get_vadjustment
				(GTK_SCROLLED_WINDOW(scrollw)));

  fview->tree = tree;

  /* Folder name column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Folder"),
						    renderer,
						    "text",
						    FV_FOLDER_COL,
						    "weight",
						    FV_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
  /* New mail count column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("New"),
						    renderer,
						    "text",
						    FV_NEW_COL,
						    "weight",
						    FV_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

  /* Unread mail count column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Unread"),
						    renderer,
						    "text",
						    FV_UNREAD_COL,
						    "weight",
						    FV_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

  /* Total mail count column */

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Total"),
						    renderer,
						    "text",
						    FV_TOTAL_COL,
						    "weight",
						    FV_WEIGHT_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

  gtk_tree_view_set_headers_visible  (GTK_TREE_VIEW(tree), TRUE );
  gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW(tree), FALSE);
  /*  gtk_tree_view_set_reorderable      (GTK_TREE_VIEW(tree), TRUE); */
  gtk_tree_view_set_search_column    (GTK_TREE_VIEW(tree), FV_FOLDER_COL);

  g_signal_connect_after(G_OBJECT(tree), 
			 "button-release-event",
			 G_CALLBACK(GtkUI_foldersview_button_release_cb),
			 fview);

  gtk_widget_show(tree);

  /* Complete XePGTKFoldersView structure */

  fview->foldersviewslist = foldersviewslist;
  fview->currentfoldersview = NULL;

  return fview;
}

static void
GtkUI_foldersview_free(XePGTKFoldersView *fview)
{
  if (!fview)
    return;

  gtk_widget_destroy(fview->widget);
  x_free(fview);

}

void
a_GtkUI_foldersview_set_list(XePGTKFoldersView *fview,
			     XePFoldersViewsList *foldersviewslist)
{
  clistiter *iter;
  GtkWidget *menu;
  GtkWidget *menuitem;
  XePFoldersView *foldersview;

  if (foldersviewslist)
    {
      menu = gtk_menu_new();
      fview->foldersviewslist = foldersviewslist;
      for (iter = clist_begin(foldersviewslist->list); iter; iter = iter->next)
	{
	  foldersview = (XePFoldersView *) iter->data;
	  menuitem = gtk_menu_item_new_with_label(foldersview->id);
	  gtk_widget_show(menuitem);
	  g_signal_connect(G_OBJECT(menuitem),
			   "activate",
			   G_CALLBACK(GtkUI_foldersview_set_cb),
			   fview);

	  gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
	}

      /* Add the autogenerated view of all mailboxes */

      foldersview = foldersviewslist->mailboxes;
      menuitem = gtk_menu_item_new_with_label(foldersview->id);
      gtk_widget_show(menuitem);
      g_signal_connect(G_OBJECT(menuitem),
		       "activate",
		       G_CALLBACK(GtkUI_foldersview_set_cb),
		       fview);
      
      gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
      
      gtk_widget_show(menu);
      gtk_option_menu_remove_menu(GTK_OPTION_MENU(fview->optmenu));
      gtk_option_menu_set_menu(GTK_OPTION_MENU(fview->optmenu), menu);
      gtk_option_menu_set_history(GTK_OPTION_MENU(fview->optmenu), 0);
      a_GtkUI_foldersview_set_view(fview, NULL);
    }
  
}

static void
GtkUI_foldersview_set_cb(GtkMenuItem *menuitem, gpointer data)
{
  XePGTKFoldersView *fview = (XePGTKFoldersView *) data;
  GtkWidget *label;

  label = gtk_bin_get_child(GTK_BIN(menuitem));

  a_GtkUI_foldersview_set_view(fview, 
			       gtk_label_get_text(GTK_LABEL(label)));
}

void
a_GtkUI_foldersview_set_view(XePGTKFoldersView *fview,
			     const char *foldersview_id)
{
  XePFoldersView *foldersview;
  
  A_LOGGING_DEBUG("Showing virtual tree %s", (foldersview_id ? foldersview_id :
					      "(NULL)"));

  foldersview = a_Foldersview_list_get_foldersview(fview->foldersviewslist, 
						   foldersview_id);
  GtkUI_foldersview_view(fview, foldersview);
}

static void
GtkUI_foldersview_view(XePGTKFoldersView *fview,
		       XePFoldersView *foldersview)
{
  cntree *tree;
  GtkTreeStore *store = fview->store;

  if (foldersview == NULL)
    return;

  tree = foldersview->tree;
  gtk_tree_store_clear(store);
  GtkUI_foldersview_append_node(tree, store, NULL);
  gtk_tree_view_expand_all(GTK_TREE_VIEW(fview->tree));
  fview->currentfoldersview = foldersview;
  GtkUI_foldersview_select_foldersviewnode(fview, foldersview->lastviewed);
}

static void
GtkUI_foldersview_set_node(GtkTreeStore *store, 
			   GtkTreeIter  *iter,
			   XePFoldersViewNode *fnode)
{
  char total[16]; /* log10(2^32) == 10, rounded up to the upper 8 boundary */
  char unseen[16];
  char recent[16];
  XePFolder *folder;
  /* Suppose folder name is not going to be larger than this */
  char fname[XEP_FOLDER_NAME_MAX_LEN]; 
  int weight;

  folder = fnode->folder_ref.folder;

  sprintf(total,  "%d", folder->total);
  sprintf(unseen, "%d", folder->unseen);
  sprintf(recent, "%d", folder->recent);

  weight = ((folder->unseen || folder->recent) ?
	    PANGO_WEIGHT_BOLD :
	    PANGO_WEIGHT_NORMAL);

  if (folder->recent)
    {
      g_snprintf(fname, sizeof(fname), "%s (%d)",
		 fnode->name ? fnode->name : a_Folder_get_id(folder),
		 folder->recent);
    }
  else
    {
      g_strlcpy(fname, fnode->name ? fnode->name : a_Folder_get_id(folder),
		sizeof(fname));
    }
 
  gtk_tree_store_set(store, iter, 
		     FV_FOLDER_COL, fname,
		     FV_NEW_COL, recent,
		     FV_UNREAD_COL, unseen,
		     FV_TOTAL_COL, total,
		     FV_WEIGHT_COL, weight,
		     FV_FOLDERSVIEWNODE_COL, fnode,
		     -1);
}

static void
GtkUI_foldersview_append_node(cntree *tree, 
			      GtkTreeStore *store, 
			      GtkTreeIter *parent)
{
  GtkTreeIter *treeiter = NULL;

  cntree *child;
  clistiter *iter;
  XePFoldersViewNode *fnode;

  if (tree == NULL)
    return;

  fnode = (XePFoldersViewNode *) tree->data;
  if (fnode)
    {
      treeiter = x_alloc(1, GtkTreeIter);
      gtk_tree_store_append(store, treeiter, parent);
      GtkUI_foldersview_set_node(store, treeiter, fnode);
    }
  if (tree->children)
    {
      for (iter = clist_begin(tree->children); iter; iter = iter->next)
	{
	  child = (cntree *) iter->data;
	  GtkUI_foldersview_append_node(child, store, treeiter);
	}
    }
  x_free(treeiter);

}

static gboolean
GtkUI_foldersview_button_release_cb(GtkWidget *widget,
				    GdkEventButton *event,
				    gpointer data)
{
  XePGTKFoldersView *fview = (XePGTKFoldersView *) data;
  GtkTreeView *treeview;
  GtkTreeIter treeiter;
  XePFoldersViewNode *fnode;
  XePFolderRef *folder_ref;
  GtkTreeModel *treemodel;
  GtkTreeSelection *selection;

  if (event->button == 1)
    {
      treeview = GTK_TREE_VIEW(fview->tree);
      selection = gtk_tree_view_get_selection(treeview);
      if (gtk_tree_selection_get_selected(selection, &treemodel, &treeiter))
	{
	  XePGTKSummaryView *summaryview;

	  gtk_tree_model_get(treemodel, &treeiter, 
			     FV_FOLDERSVIEWNODE_COL, &fnode, -1);
	  folder_ref = &(fnode->folder_ref);

	  summaryview = (XePGTKSummaryView *) fview->uimain->summaryview;
	  a_GtkUI_summaryview_set_view_async(summaryview, folder_ref);

	  fview->currentfoldersview->lastviewed = fnode;
	}
      return TRUE;
    }
  return TRUE;
}

static void
GtkUI_foldersview_select_foldersviewnode(XePGTKFoldersView *fview,
					 XePFoldersViewNode *fnode)
{

  if (GtkUI_foldersview_select_node(fview, fnode, NULL))
    {
      XePGTKSummaryView *summaryview;

      summaryview = (XePGTKSummaryView *) fview->uimain->summaryview;

      fview->currentfoldersview->lastviewed = fnode;

      a_GtkUI_summaryview_set_view_async(summaryview, &(fnode->folder_ref));
    }
}

static gboolean
GtkUI_foldersview_select_node(XePGTKFoldersView *fview, 
			      XePFoldersViewNode *fnode,
			      GtkTreeIter *treeiter)
{
  gboolean valid;
  GtkTreeIter child;
  GtkTreeModel *treemodel;
  GtkTreeView *treeview = GTK_TREE_VIEW(fview->tree);
  
  treemodel = gtk_tree_view_get_model(treeview);

  valid = gtk_tree_model_iter_children(treemodel, &child, treeiter);
  while (valid)
    {
      XePFoldersViewNode *afoldersviewnode;

      gtk_tree_model_get(treemodel, &child, 
			 FV_FOLDERSVIEWNODE_COL, &afoldersviewnode, -1);
      if (afoldersviewnode == fnode)
	{
	  GtkTreeSelection *selection;
	  selection = gtk_tree_view_get_selection(treeview);
	  gtk_tree_selection_select_iter(selection, &child);
	  valid = FALSE;
	  return TRUE;
	}
      else 
	{
	  if (gtk_tree_model_iter_has_child(treemodel, &child))
	    GtkUI_foldersview_select_node(fview, fnode, &child);
	  valid = gtk_tree_model_iter_next(treemodel, &child);
	}
    }
  return FALSE;
}

static gboolean
GtkUI_foldersview_refresh_node(GtkTreeModel *model,
			       GtkTreePath  *path,
			       GtkTreeIter  *iter,
			       gpointer      data)
{
  XePFoldersViewNode *fnode;

  gtk_tree_model_get(model, iter, FV_FOLDERSVIEWNODE_COL, &fnode, -1);
  
  GtkUI_foldersview_set_node(GTK_TREE_STORE(model), iter, fnode);
  return FALSE;
}

static void
GtkUI_foldersview_refresh(XePGTKFoldersView *fview)
{
  GtkTreeModel *treemodel;
  GtkTreeView *treeview = GTK_TREE_VIEW(fview->tree);
  
  treemodel = gtk_tree_view_get_model(treeview);
  
  gtk_tree_model_foreach(treemodel,
			 GtkUI_foldersview_refresh_node,
			 fview);
}

/* Summary View */
static XePGTKSummaryView *
GtkUI_summaryview_new(XePUIMain *uimain)
{
  XePGTKSummaryView *sview;
  GtkWidget *vbox;
  GtkWidget *tree;
  GtkWidget *image;

  sview = x_alloc(1, XePGTKSummaryView);
  if (!sview)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }

  sview->uimain = uimain;
  sview->folder = NULL;
  sview->pane_position = 0;

  sview->widget = vbox = gtk_vbox_new(FALSE, 0);

  /* The real summary view widge */

  sview->tree = tree = gtk_xtree_view_new();
  gtk_box_pack_start(GTK_BOX(vbox), tree, TRUE, TRUE, 0);

  /* New */

  image = gtk_image_new_from_stock(GTK_STOCK_YES, GTK_ICON_SIZE_MENU);
  gtk_widget_show(image);
  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_STATUS_COL,
			       GTK_XCOLUMN_TYPE_STOCKID, 
			       TRUE, GTK_SORT_ASCENDING, image, NULL);
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_STATUS_COL, 3*8);
			       
  /* Flagged */
  image = gtk_image_new_from_stock("xetpan-flag", GTK_ICON_SIZE_MENU);
  gtk_widget_show(image);
  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_FLAG_COL,
			       GTK_XCOLUMN_TYPE_STOCKID, 
			       TRUE, GTK_SORT_ASCENDING, image, NULL);
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_FLAG_COL, 3*8);

  /* Subject column */

  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_SUBJECT_COL,
			       GTK_XCOLUMN_TYPE_TEXT, 
			       TRUE, GTK_SORT_ASCENDING, NULL, _("Subject"));
  
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_SUBJECT_COL, 14*8);
  gtk_xtree_view_set_expander_column(GTK_XTREE_VIEW(tree), SV_SUBJECT_COL);
  gtk_xtree_view_set_expansible_column(GTK_XTREE_VIEW(tree), SV_SUBJECT_COL);


  /* From column */

  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_FROM_COL,
			       GTK_XCOLUMN_TYPE_TEXT, 
			       TRUE, GTK_SORT_ASCENDING, NULL, _("From"));
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_FROM_COL, 14*8);
  
  /* Date column */

  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_DATE_COL,
		       GTK_XCOLUMN_TYPE_TEXT, 
			       TRUE, GTK_SORT_ASCENDING, NULL, _("Date"));
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_DATE_COL, 14*8);

  /* Size column */

  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_SIZE_COL,
			       GTK_XCOLUMN_TYPE_TEXT, 
			       TRUE, GTK_SORT_ASCENDING, NULL, _("Size"));
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_SIZE_COL, 5*8);

  /* Data function */
#ifdef VIEW_TREE
  gtk_xtree_view_set_row_data_func(GTK_XTREE_VIEW(tree), 
				   GtkUI_summaryview_tree_data_func,
				   sview);

  gtk_xtree_view_set_row_params_func(GTK_XTREE_VIEW(tree), 
				     GtkUI_summaryview_tree_params_func,
				     sview);
  gtk_xtree_view_set_data_iterator_func(GTK_XTREE_VIEW(tree), 
					GtkUI_summaryview_tree_iterator_func,
					sview);
#else
  gtk_xtree_view_set_row_data_func(GTK_XTREE_VIEW(tree), 
				   GtkUI_summaryview_list_data_func,
				   sview);

  gtk_xtree_view_set_row_params_func(GTK_XTREE_VIEW(tree), 
				     GtkUI_summaryview_list_params_func,
				     sview);
  gtk_xtree_view_set_data_iterator_func(GTK_XTREE_VIEW(tree), 
					GtkUI_summaryview_list_iterator_func,
					sview);
#endif
  g_signal_connect_after(G_OBJECT(tree), 
			 "button-release-event",
			 G_CALLBACK(GtkUI_summaryview_button_release_cb),
			 sview);

  gtk_widget_show(tree);

  return sview;
}

static void
GtkUI_summaryview_free(XePGTKSummaryView *sview)
{
  if (!sview)
    return;

  gtk_widget_destroy(sview->widget);
  x_free(sview);
}

static gboolean
GtkUI_summaryview_tree_data_func(void *iterator, 
				 guint col_id,
				 void **content,
				 gchar **attribute,
				 gpointer data)
{
  struct mail_envelopes_tree *msg_tree;
  mailmessage *msg = NULL;

  msg_tree =  (struct mail_envelopes_tree *) iterator;
  msg = msg_tree->msg;

  return GtkUI_summaryview_msg_data_func(msg, 
					 col_id, 
					 content, 
					 attribute, 
					 data);
}

static gboolean
GtkUI_summaryview_msg_data_func(mailmessage *msg, 
				guint col_id,
				void **content,
				gchar **attribute,
				gpointer data)
{
  static char buf[1024];
  char *dsp_subject = NULL;
  char *dsp_from = NULL;
  char *display_name = NULL;
  char *email_address = NULL;
  char *msg_date = NULL;
  const char *unit = "bkMGTPEZY?";
  size_t size;

  if (msg == NULL)
    {
      *content = NULL;
      *attribute = NULL;
      return TRUE;
    }

  switch (col_id)
    {
    case SV_STATUS_COL:
      *content = (msg->flags->flags & MAIL_FLAG_DELETED ?
		  GTK_STOCK_DELETE :
		  msg->flags->flags & MAIL_FLAG_NEW ?
		  GTK_STOCK_YES :
		  !(msg->flags->flags & MAIL_FLAG_SEEN) ?
		  GTK_STOCK_NO :
		  msg->flags->flags & MAIL_FLAG_ANSWERED ?
		  "xetpan-replied" :
		  msg->flags->flags & MAIL_FLAG_FORWARDED ?
		  "xetpan-forwarded" :
		  NULL);
      break;
    case SV_FLAG_COL:
      *content =  ((msg && msg->flags && 
		    (msg->flags->flags & MAIL_FLAG_FLAGGED)) ?
		   "xetpan-flag" : NULL);
	break;
    case SV_SUBJECT_COL:
      a_Message_message_get_subject(&msg->single_fields, 
				    GTK_DEST_CHARSET,
				    &dsp_subject);
      if (dsp_subject)
	{
	  strncpy(buf, dsp_subject, sizeof(buf) - 1);
	  x_free(dsp_subject);
	  *content = buf;
	}
      else
	{
	  *content = NULL;
	}
      break;
    case SV_FROM_COL:
      a_Message_message_get_from(&msg->single_fields, GTK_DEST_CHARSET,
				 &display_name, &email_address);
      if (display_name)
	{
	  dsp_from = display_name;
	  x_free(email_address);
	}
      else if (email_address)
	{
	  dsp_from = email_address;
	}
      if (dsp_from)
	{
	  strncpy(buf, dsp_from, sizeof(buf) - 1);
	  x_free(dsp_from);
	  
	  *content = buf;
	}
      else
	{
	  *content = NULL;
	}
      break;
    case SV_DATE_COL:
      msg_date = 
	a_Message_message_formatted_get_local_date
	(&msg->single_fields, "%R %d-%m-%y", GTK_DEST_CHARSET);
      if (msg_date)
	{
	  strncpy(buf, msg_date, sizeof(buf) - 1);
	  x_free(msg_date);
	  *content = buf;
	}
      else
	{
	  *content = NULL;
	}

      break;
    case SV_SIZE_COL
:      /* b: byte
       * k: kilo byte = 1024 ~ 1e3
       * M: Mega byte = 1024*1024 ~ 1e6
       * G: Giga byte = 1024*1024*1024 ~ 1e9
       * P: Peta byte = 1024*1024*1024*1024 ~ 1e12
       * E: Exa byte  = 1024*1024*1024*1024*1024 ~ 1e15
       * Z: Zetta byte= 1024*1024*1024*1024*1024*1024 ~ 1e18
       * Y: Yotta byte= 1024*1024*1024*1024*1024*1024*1024 ~ 1e21 !
       */
      size = msg->size;
      for (; *unit && size > (1 << 10); unit++)
	{
	  size = size >> 10;
	}
      sprintf(buf, "%4d%c", size, *unit);
      *content = buf;
      break;

    }

  if (msg->flags && ((msg->flags->flags & MAIL_FLAG_NEW) ||
		     !(msg->flags->flags & MAIL_FLAG_SEEN)))
    {
      *attribute = "bold";
    }
  else
    {
      *attribute = NULL;
    }
  return TRUE;
}

static void
GtkUI_summaryview_tree_params_func(void *iterator, 
				   gboolean *has_children,
				   guint *depth,
				   gpointer data)
{
  struct mail_envelopes_tree *msg_tree;
  int i = 0;

  msg_tree = (struct mail_envelopes_tree *) iterator;

  *has_children = (msg_tree->children && carray_count(msg_tree->children) ? 
		   TRUE :
		   FALSE);
  while (msg_tree->parent)
    {
      msg_tree = msg_tree->parent;
      i++;
    }
  /* Do not count the root, i.e. root's children are at depth 0 */
  i--;
  *depth = i;
}

static gboolean
GtkUI_summaryview_tree_iterator_func(void **iterator, 
				     gboolean find_next,
				     gboolean ignore_children,
				     guint sort_col_id,
				     GtkSortType sort_order,
				     gpointer data)
{
  struct mail_envelopes_tree *msg_tree;
  gboolean found = FALSE;

  msg_tree = (struct mail_envelopes_tree *) *iterator;
  
  if (find_next)
    {
      do {
	msg_tree = mail_envelopes_tree_next(msg_tree, NULL, ignore_children);
	if (msg_tree)
	  found = TRUE; /* make it a function to filtered out items */
      } while (msg_tree && !found);
    }
  else
    {
      do {
	msg_tree = mail_envelopes_tree_prev(msg_tree->parent, msg_tree);
	if (msg_tree)
	  found = TRUE;  /* make it a function to filtered out items */
      } while (msg_tree && !found);
    }

  if (found) 
    *iterator = msg_tree;
  return found;
}


static struct mail_envelopes_tree *
mail_envelopes_tree_next(struct mail_envelopes_tree *self,
			 struct mail_envelopes_tree *child,
			 int ignore_children)
{
  uint32_t len;

  if (self == NULL)
    return NULL;

  if (!ignore_children && 
      self->children && 
      (len = carray_count(self->children)))
    {
      carray *children = self->children;
      uint32_t i;
      
      if (child == NULL)
	return carray_get(children, 0);

      for (i = 0; i < len && carray_get(children, i) != child; i++)
	{
	  /* Empty on purpose */
	}
      if (i < len - 1)
	return carray_get(children, i + 1);
    }

  return mail_envelopes_tree_next(self->parent, self, 0);
}

static struct mail_envelopes_tree *
mail_envelopes_tree_prev(struct mail_envelopes_tree *parent,
			 struct mail_envelopes_tree *self)
{
  carray *siblings;
  struct mail_envelopes_tree *prev;
  uint32_t len;
  uint32_t i;
  
  if (parent == NULL)
    return NULL;

  siblings = parent->children;
  len = carray_count(siblings);

  for (i = 0; i < len && carray_get(siblings, i) != self; i++)
    {
      /* Empty on purpose */
    }
  if (i > 0)
    {
      /* find last descendent */
      prev = carray_get(siblings, i - 1);
      while (prev->children->len)
	{
	  prev = carray_get(prev->children, prev->children->len - 1);
	};
      return prev;
    }
  /* if parent is the root, we reached the top */
  if (parent && parent->parent)
    return parent;
  else
    return NULL;
}

static gboolean
GtkUI_summaryview_list_data_func(void *iterator, 
				 guint col_id,
				 void **content,
				 gchar **attribute,
				 gpointer data)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) data;
  struct mail_envelopes_list *ep_list = sview->folder->ep_list;
  uint32_t index;
  mailmessage *msg = NULL;

  index = (uint32_t) iterator;
  if (index) 
    {
      /* iterator 0 is an invalid one */
      /* iterator start from 1 but array indexes must start from 0
       * so we adjust index to point into the array 
       */
      msg = carray_get(ep_list->tab, index - 1);
    }
  return GtkUI_summaryview_msg_data_func(msg,
					 col_id,
					 content, 
					 attribute, 
					 data);
}

static void
GtkUI_summaryview_list_params_func(void *iterator, 
				   gboolean *has_children,
				   guint *depth,
				   gpointer data)
{
  /* list item do not have any children and are at depth 0 */
  *has_children = FALSE;
  *depth = 0;

}

static gboolean
GtkUI_summaryview_list_iterator_func(void **iterator, 
				     gboolean find_next,
				     gboolean ignore_children,
				     guint sort_col_id,
				     GtkSortType sort_order,
				     gpointer data)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) data;
  struct mail_envelopes_list *ep_list = sview->folder->ep_list;
  gboolean found = FALSE;
  uint32_t index = (uint32_t) *iterator;

  /*  printf("Finding %s iterator to %d\n", (find_next ? "next" : "prev"), index);*/
  if (find_next)
    {
      do
	{
	  index = mail_envelopes_list_next(ep_list, index);
	  if (index)
	    found = TRUE; /* change to a function to filter out items */
	  
	} while (index && !found);
    }
  else
    {
      do
	{
	  index = mail_envelopes_list_prev(ep_list, index);
	  if (index)
	    found = TRUE;
	} while (index && !found);
    }
  if (found)
    *iterator = (void *) index;

  /*  printf("...found: %d\n", index);*/

  return found;
}

static uint32_t
mail_envelopes_list_next(struct mail_envelopes_list *ep_list, uint32_t index)
{
  /* index starts from 1 but array index start from 0 */
  if (index < carray_count(ep_list->tab))
    return index + 1;
  else
    return 0;
}

static uint32_t
mail_envelopes_list_prev(struct mail_envelopes_list *ep_list, uint32_t index)
{
  /* index starts from 1 but array index start from 0 */
  if (index - 1 > 0)
    return index - 1;
  else
    return 0;
}

static mailmessage *
GtkUI_summaryview_get_selected_msg(XePGTKSummaryView *sview)
{
  struct mail_envelopes_tree *msg_tree;

  msg_tree = gtk_xtree_view_get_focused(GTK_XTREE_VIEW(sview->tree));

  if (msg_tree)
    return msg_tree->msg;
  else
    return NULL;
}

static gboolean
GtkUI_summaryview_button_release_cb(GtkWidget *widget,
				    GdkEventButton *event,
				    gpointer data)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) data;
  XePGTKMessageView *mview  = (XePGTKMessageView *) sview->uimain->messageview;
  XePUIMain *uimain = sview->uimain;
  int col_id;

  if (event->button == 1 || event->button == 2)
    {
      mailmessage *msg;
      XePMessage *message;
      msg = GtkUI_summaryview_get_selected_msg(sview);
      if (msg)
	{
	  if (gtk_xtree_view_find_cell(GTK_XTREE_VIEW(sview->tree),
				       event->x, event->y,
				       NULL,
				       NULL, &col_id))
	    {
	      if (col_id == SV_FLAG_COL)
		{
		  if (msg->flags)
		    {
		      msg->flags->flags ^= MAIL_FLAG_FLAGGED;
		      gtk_xtree_view_update(GTK_XTREE_VIEW(sview->tree));
		    }
		  return TRUE;
		}
	    }
	  if (event->button == 2)
	    {
	      gchar *display;
	      display = a_GtkUtils_make_display_name(uimain->toplevel);
	      mview = 
		(XePGTKMessageView *) a_UI_messageview_create(uimain->xetpan,
							      display, 
							      uimain->type);
	      g_free(display);
	    }

	  message = a_Message_new_for_fetching(sview->folder, msg);
	  if (message)
	    {
	      a_UI_messageview_show_message((XePUIMessageView *)mview, 
					    message);
	    }

	}
      return TRUE;
    }
  return TRUE;
}

void
a_GtkUI_summaryview_set_view_async(XePGTKSummaryView *sview,
				   XePFolderRef *folder_ref)
{
  if (folder_ref)
    {
      gchar *toplevel_title;

      A_LOGGING_DEBUG("viewing summary of %s\n", folder_ref->ref);

      toplevel_title = g_strdup_printf(_("XetPan - %s"), folder_ref->ref);
      gtk_window_set_title(GTK_WINDOW(sview->uimain->toplevel),
			   toplevel_title);
      g_free(toplevel_title);
      sview->folder = folder_ref->folder;
      gtk_xtree_view_freeze(GTK_XTREE_VIEW(sview->tree));
      a_UI_update_folder(folder_ref->ref);

    }
  else 
    {
      A_LOGGING_SYSTEM_ERROR("FolderID == NULL");
      sview->folder = NULL;
    }
}

static void
GtkUI_summaryview_delete_item(XePGTKSummaryView *sview, mailmessage *msg)
{
  XePMessage *message;
  XePManagerMsg *managermsg;

  /* FIXME: all views having this message must be updated */

  if (msg)
    {
      /* FIXME: here we suppose that the delete toggling will invariably
       * succeed
       */
      message = a_Message_new();
      message->folder = sview->folder;
      message->msg = msg;
      message->flags_set = (msg->flags->flags & MAIL_FLAG_DELETED ? 0 : MAIL_FLAG_DELETED);
      message->flags_unset = (msg->flags->flags & MAIL_FLAG_DELETED ? MAIL_FLAG_DELETED : 0);
       
      managermsg = a_Manager_msg_new();
      if (managermsg == NULL)
	{
	  a_Message_free(message);
	  return;
	}
      
      a_Manager_msg_set_type(managermsg, XEP_MANAGER_FLAGS_CHANGE);
      a_Manager_msg_set_source_target(managermsg,
				      NULL,
				      &message->folder->thread);
      a_Manager_msg_set_args(managermsg, message, message);
      a_Manager_msg_add_callback(managermsg, a_Message_free_cb, message);
      a_Manager_msg_add_callback(managermsg, GtkUI_summaryview_update_cb, sview);
      a_Manager_msg_submit(managermsg);
    }
}

static void GtkUI_summaryview_update_cb(void *asview, void *amessage)
{
  /* We should update only one row */
  XePGTKSummaryView *sview = (XePGTKSummaryView *) asview;
  gtk_xtree_view_update(GTK_XTREE_VIEW(sview->tree));
}

void
a_GtkUI_summaryview_delete_selection(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  void **selection;
  uint32_t n;
  mailmessage *msg;

  n = gtk_xtree_view_get_selection(GTK_XTREE_VIEW(sview->tree), &selection);

  if (n)
    {
      uint32_t i;
      for (i = 0; i < n; i++)
	{
	  msg = ((struct mail_envelopes_tree *) selection[i])->msg;
	  GtkUI_summaryview_delete_item(sview, msg);
	}
    }
}

void
a_GtkUI_summaryview_expunge(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  XePFolder *folder = sview->folder;

  gtk_xtree_view_freeze(GTK_XTREE_VIEW(sview->tree));
  a_UI_expunge_folder(folder->id);
}

void
a_GtkUI_folder_updated(XePUIMain *uimain, XePFolder *folder)
{

  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  A_LOGGING_DEBUG("folder '%s'\n", a_Folder_get_id(folder));

  if (sview->folder == folder)
    {
      GtkUI_summaryview_refresh(sview);
    }

  /* Update folderview node */

  GtkUI_foldersview_refresh((XePGTKFoldersView *)(uimain->foldersview));
}

static void
GtkUI_summaryview_refresh(XePGTKSummaryView *sview)
{
  XePFolder *folder = sview->folder;
#ifdef XETPAN_TIMES
  struct tms tic, tac, toc, tuc;

#define LAP(tic, tac) \
    printf("user = %.4e sys = %.4e " \
	   "chdu = %.4e chds= %.4e\n", \
	   ((float)(tac.tms_utime - tic.tms_utime))/CLOCKS_PER_SEC, \
	   ((float)(tac.tms_stime - tic.tms_stime))/CLOCKS_PER_SEC, \
	   ((float)(tac.tms_cutime - tic.tms_cutime))/CLOCKS_PER_SEC, \
	   ((float)(tac.tms_cstime - tic.tms_cstime))/CLOCKS_PER_SEC)
	   
#endif

  a_GtkMessageview_message_clear((XePGTKMessageView*)sview->uimain->messageview);

#ifdef XETPAN_TIMES
  times(&tic);
#endif
  
  printf("clearing...\n");

#ifdef XETPAN_TIMES
  times(&tac);
#endif

  printf("filling...\n");

#ifdef VIEW_TREE
  gtk_xtree_view_set_first_row(GTK_XTREE_VIEW(sview->tree), carray_get(folder->ep_tree->children, 0));
  {
    struct mail_envelopes_tree *msg_tree = carray_get(folder->ep_tree->children, 0);
    printf("First child %p msg %p parent %p\n", msg_tree, msg_tree->msg, msg_tree->parent);
  }
#else
  gtk_xtree_view_set_first_row(GTK_XTREE_VIEW(sview->tree), (void *) 1);
#endif
  gtk_xtree_view_set_length(GTK_XTREE_VIEW(sview->tree), folder->total);
  printf("total messages = %d\n", folder->total);

  gtk_xtree_view_thaw(GTK_XTREE_VIEW(sview->tree));

  printf("setting...\n");

#ifdef XETPAN_TIMES
  times(&toc);
#endif
  printf("expanding...\n");

#ifdef XETPAN_TIMES
  times(&tuc);

  printf("Clearing: ");
  LAP(tic, tac);
  
  printf("Filling:  ");
  LAP(tac, toc);

  printf("Expanding:");
  LAP(toc, tuc);
#endif
}

static void
GtkUI_summaryview_update_selected(XePGTKSummaryView *sview)
{
  gtk_xtree_view_update(GTK_XTREE_VIEW(sview->tree));
}

/* Message View */
XePUIMessageView *
a_GtkUI_messageview_create_standalone(XetPan *xetpan, 
				      char *terminal)
{
  return (XePUIMessageView *) a_GtkMessageview_create_standalone(xetpan, 
								 terminal, 
								 0);
}

void
a_GtkUI_messageview_close(XePUIMessageView *amview)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  A_LOGGING_TRACE;
  a_GtkMessageview_free(mview);
}

void
a_GtkUI_messageview_finalize(XePUIMessageView *amview, XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  if (mview->uimain && mview->uimain->summaryview)
    {
      XePGTKSummaryView *sview = 
	(XePGTKSummaryView *) mview->uimain->summaryview;

      GtkUI_summaryview_update_selected(sview);

    }
  a_GtkMessageview_finalize((XePGTKMessageView *) mview, message);
}

void
a_GtkUI_messageview_mime_view_show(XePUIMessageView *amview,
				   XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;
  if (mview->muid != message->muid)
    return;

  a_GtkMessageview_mime_view_show((XePGTKMessageView *) mview, message);
}

void
a_GtkUI_messageview_mime_view_block(XePUIMessageView *amview,
				    XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  if (mview->muid != message->muid)
    return;

  a_GtkMessageview_mime_view_block((XePGTKMessageView *) mview, message);
}

void
a_GtkUI_messageview_mime_view_unblock(XePUIMessageView *amview, 
				      XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  if (mview->muid != message->muid)
    return;

  a_GtkMessageview_mime_view_unblock((XePGTKMessageView *) mview, message);
}

void
a_GtkUI_messageview_mime_show_selected(XePUIMessageView *amview, 
				       XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  if (mview->muid != message->muid)
    return;

  a_GtkMessageview_mime_show_selected((XePGTKMessageView *) mview, message);
}

/* Compose Window */
XePUIComposeWindow *
a_GtkUI_composewindow_create(XetPan *xetpan, char *terminal)
{
  return (XePUIComposeWindow *) a_GtkComposewindow_create(xetpan, terminal);
}

void
a_GtkUI_composewindow_set_message(XePUIComposeWindow *uicompose, 
				  XePMessage *message)
{
  a_GtkComposewindow_set_message((XePGTKComposeWindow *) uicompose, message);
}

void
a_GtkUI_composewindow_set_body_text(XePUIComposeWindow *uicompose,
				    const char *text,
				    size_t text_size,
				    const char *text_charset)
{
  a_GtkComposewindow_set_body_text((XePGTKComposeWindow *) uicompose, 
				   text,
				   text_size,
				   text_charset);
}

void
a_GtkUI_composewindow_add_header(XePUIComposeWindow *uicompose, 
				 XePMessageHeaderType header_type,
				 const char *header_name,
				 const char *header_value)
{
  a_GtkComposewindow_add_header((XePGTKComposeWindow *) uicompose,
				header_type,
				header_name,
				header_value);
}

void
a_GtkUI_composewindow_close(XePUIComposeWindow *uicompose)
{
  A_LOGGING_TRACE;
  a_GtkComposewindow_free((XePGTKComposeWindow *) uicompose);
}

/*************************************************************************/
/*                                                                       */
/* Watching IO traffic                                                   */
/*                                                                       */
/*************************************************************************/

int
a_GtkUI_create_io_watch(XePUI *ui, XePUIIOWatcher *watcher)
{
  GIOChannel *channel;
  GIOCondition gcondition = 0;
  short int condition = watcher->condition;
  unsigned int tag;

  A_LOGGING_TRACE;

  if (condition & XETPAN_POLLIN)
    gcondition |= G_IO_IN;
  if (condition & XETPAN_POLLOUT)
    gcondition |= G_IO_OUT;
  if (condition & XETPAN_POLLPRI)
    gcondition |= G_IO_PRI;
  if (condition & XETPAN_POLLERR)
    gcondition |= G_IO_ERR;
  if (condition & XETPAN_POLLHUP)
    gcondition |= G_IO_HUP;
  if (condition & XETPAN_POLLNVAL)
    gcondition |= G_IO_NVAL;
  
  channel = g_io_channel_unix_new(watcher->fd);
  
  g_io_channel_set_encoding(channel, NULL, NULL);
  
  tag = g_io_add_watch(channel, gcondition, GtkUI_io_watch_cb, watcher);
  
  watcher->channel = channel;
  watcher->tag = tag;
  
  return 0;
}

void
a_GtkUI_delete_io_watch(XePUI *ui, XePUIIOWatcher *watcher)
{
  A_LOGGING_DEBUG("deleting fd=%d", watcher->fd);

  g_source_remove(watcher->tag);
  g_io_channel_unref((GIOChannel *) watcher->channel);
}

static gboolean
GtkUI_io_watch_cb(GIOChannel *channel, GIOCondition gcondition, gpointer data)
{
  XePUIIOWatcher *watcher = (XePUIIOWatcher *) data;
  int condition = 0;

  if (gcondition & G_IO_IN)
    condition |= XETPAN_POLLIN;
  if (gcondition & G_IO_OUT)
    condition |= XETPAN_POLLOUT;
  if (gcondition & G_IO_PRI)
    condition |= XETPAN_POLLPRI;
  if (gcondition & G_IO_ERR)
    condition |= XETPAN_POLLERR;
  if (gcondition & G_IO_HUP)
    condition |= XETPAN_POLLHUP;
  if (gcondition & G_IO_NVAL)
    condition |= XETPAN_POLLNVAL;

  watcher->callback(watcher->fd, condition, watcher->data);
  return TRUE;
}

#endif /* USE_GTKUI */
