/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This is part of XetPan
 *
 * This is a tree view/list view widget taylored for small memory footprint
 * and for direct accessing of user data via user provided function
 * to retrieve data and to navigate through it.
 * 
 * Some parts of this code are heavily inspired from gtk+ 2.x gtktreeview,
 * for example: resize grips in header button, some function names.
 *
 * Gtk+ is to be found at http://www.gtk.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtksignal.h>
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtktable.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkviewport.h>
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkarrow.h>
#include <gtk/gtkhscrollbar.h>
#include <gtk/gtkvscrollbar.h>
#include <gtk/gtkrange.h>
#include <gtk/gtktreeview.h>

#include "gtkxtreeview.h"

#include <math.h>

#define H_SPACE 12

#undef DEBUG

enum {
	GTK_XTREE_VIEW_HAS_CHILDREN = 1 << 31,
	GTK_XTREE_VIEW_COLLAPSED    = 1 << 30,
	GTK_XTREE_VIEW_SELECTED     = 1 << 29,
	GTK_XTREE_VIEW_DEPTH = ~(GTK_XTREE_VIEW_SELECTED | 
			GTK_XTREE_VIEW_COLLAPSED |
			GTK_XTREE_VIEW_HAS_CHILDREN)
};
#define g_marshal_value_peek_int(v)     (v)->data[0].v_int
#define g_marshal_value_peek_uint(v)     (v)->data[0].v_uint

#define DAREA_DEFAULT_WIDTH 100
#define DAREA_DEFAULT_HEIGHT 100
typedef struct _GtkXTreeViewColumn GtkXTreeViewColumn;

struct _GtkXTreeViewColumn
{
	guint col_id;

	GtkXColumnType type;

	gboolean is_sortable;
	GtkSortType sort_order;

	guint width;

	GtkWidget *widget; /* The button of the title */
	GtkWidget *hbox; /* the hbox in the title button */
	GtkWidget *user_widget; /* The user provided widget for the title */
	GtkWidget *label; /* the label of the title button */
	GtkWidget *arrow; /* the sorting direction arrow in the title button */

	GtkXTreeView *treeview;

	GdkWindow *resize_grip; /* to resize */
	gboolean being_resized;
};

struct _GtkXTreeViewClass
{
	GtkTableClass parent_class;

	void (*sort_changed)(GtkXTreeView *xtv, guint col_id, GtkSortType sort_type, gpointer data); /* default handler */
};

struct _GtkXTreeView
{
	GtkTable table;

	GtkWidget *vbox;
	GtkWidget *hbox; /* hbox containing title buttons */
	GtkWidget *darea; /* drawing area */
	GtkWidget *viewport; /* viewport */
	GtkWidget *hscrollbar; /* horizontal scrolling */
	GtkWidget *vscrollbar; /* scrolls up and down, controls first row iterator */

	guint col_nb;
	guint sort_col_id;
	GtkSortType sort_order;
	guint expander_col_id;
	guint expander_width;
	guint expansible_col_id;

	gboolean enabled;

	void *first_iterator;
	void *last_iterator;
	guint last_row;
	guint first_value;

	void *focused_iterator;
	uint32_t total_selected;

	GSList *columns; /* items are GtkXTreeViewColumn* */

	guint allocated_rows;
	guint allocated_cols;

	void ***cells; /* cells[0..rows][0..cols] are pointer to PangoLayout or 
				Pixbufs */
	guint *cell_prop;
	guint *cell_prop_tmp;
	void **iterators;
	void **iterators_tmp;
	gint *y_pos;
	gint *y_pos_tmp;
	guint rows;
	guint length;

	/* User callback functions */
	GtkXTreeViewRowDataFunc row_data_func;
	gpointer row_data_data;

	GtkXTreeViewRowParamsFunc row_params_func;
	gpointer row_params_data;

	GtkXTreeViewDataIteratorFunc data_iterator_func;
	gpointer data_iterator_data;
};

enum {
	SORT_CHANGED_SIGNAL,
	LAST_SIGNAL
};

static guint gtkxtreeview_signals[LAST_SIGNAL] = {0};


/**/

static void
_g_cclosure_marshal_VOID__UINT_INT (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data);

static void gtk_xtree_view_class_init(GtkXTreeViewClass *klass);
static void gtk_xtree_view_init(GtkXTreeView *xtv);
static void darea_size_request_cb(GtkWidget *widget, 
		GtkRequisition *requisition);
static void darea_size_allocate_cb(GtkWidget *widget, 
		GtkAllocation *allocation, 
		gpointer axtv);
static gboolean darea_configure_event_cb(GtkWidget *widget, 
		GdkEventConfigure *event, 
		gpointer axtv);
static gboolean darea_expose_event_cb(GtkWidget *widget, 
		GdkEventExpose *event, 
		gpointer axtv);
static void scroll_cb(GtkRange *range, 
		gpointer axtv);

static void darea_fill(GtkXTreeView *xtv, 
		void *firs_iterator, 
		guint first_row, 
		guint last_row);
static void cache_set_row_parameter(GtkXTreeView *xtv, void *cache_iterator, 
		gboolean has_children, 
		guint depth);
static void cache_set_node_status(GtkXTreeView *xtv, void *cache_iterator, 
		gboolean is_selected, 
		gboolean is_collapsed);
static void cache_get_node_status(GtkXTreeView *xtv, void *cache_iterator, 
		gboolean *is_selected, 
		gboolean *is_collapsed);
static void add_row(GtkXTreeView *xtv);
static void *cache_get_row_iterator(GtkXTreeView *xtv, void *iterator);
static void create_title_button(GtkXTreeViewColumn *column, 
		GtkWidget *title_widget, 
		gchar *title_string);

static void gtk_xtree_view_update_adjustment(GtkXTreeView *xtv);
static void update_resize_grips(GtkXTreeView *xtv);
static void header_realize_cb(GtkWidget *button, 
		gpointer acolumn);
static void header_clicked_cb(GtkButton *button, 
		gpointer acolumn);
static gboolean header_button_press_cb(GtkWidget *widget, 
		GdkEventButton *event, gpointer acolumn);
static gboolean header_button_release_cb(GtkWidget *widget, 
		GdkEventButton *event, 
		gpointer column);
static gboolean header_event_cb(GtkWidget *widget, 
		GdkEvent *event, 
		gpointer acolumn);

static gboolean darea_button_release_cb(GtkWidget *widget, 
		GdkEventButton *event,
		gpointer axtv);
static gboolean darea_button_press_cb(GtkWidget *widget, 
		GdkEventButton *event,
		gpointer axtv);

static void
_g_cclosure_marshal_VOID__UINT_INT (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data)
{
  typedef void (*GMarshalFunc_VOID__UINT_INT) (gpointer     data1,
		  			   guint        arg_1,
                                           gint         arg_2,
                                           gpointer     data2);
  register GMarshalFunc_VOID__UINT_INT callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_VOID__UINT_INT) (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_uint (param_values + 1),
            g_marshal_value_peek_int (param_values + 2),
            data2);
}

GType gtk_xtree_view_get_type(void)
{
	static GType xtv_type = 0;

	if (!xtv_type) {
		static const GTypeInfo xtv_info = 
		{
			sizeof(GtkXTreeViewClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) gtk_xtree_view_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof(GtkXTreeView ),
			0, /* n_preallocs */
			(GInstanceInitFunc) gtk_xtree_view_init
		};

		xtv_type = g_type_register_static(GTK_TYPE_TABLE, "GtkXTreeView", &xtv_info, 0);
	}

	return xtv_type;
}

static void gtk_xtree_view_class_init(GtkXTreeViewClass *klass)
{
	gtkxtreeview_signals[SORT_CHANGED_SIGNAL] = g_signal_new("sort-changed",
			G_TYPE_FROM_CLASS(klass),
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			G_STRUCT_OFFSET(GtkXTreeViewClass, sort_changed),
			NULL, NULL,
			_g_cclosure_marshal_VOID__UINT_INT,
			G_TYPE_NONE,
			2,
			G_TYPE_UINT,
			G_TYPE_INT);
}

static void gtk_xtree_view_init(GtkXTreeView *xtv)
{
	gint attach;
	GtkStyle *style;

	xtv->col_nb = 0;
	xtv->sort_col_id = 0;
	xtv->sort_order = GTK_SORT_ASCENDING;
	xtv->expander_col_id = 0;
	xtv->expander_width = 10;
	xtv->expansible_col_id = 0;
	xtv->enabled = FALSE;
	xtv->first_iterator = NULL;
	xtv->last_iterator = NULL;
	xtv->last_row = 0;
	xtv->first_value = 0;

	xtv->focused_iterator = NULL;
	xtv->total_selected = 0;

	xtv->columns = NULL;

	xtv->allocated_rows = 0;
	xtv->allocated_cols = 0;

	xtv->cells = NULL;
	xtv->cell_prop = NULL;
	xtv->cell_prop_tmp = NULL;
	xtv->y_pos = NULL;
	xtv->y_pos_tmp = NULL;
	xtv->iterators = NULL;
	xtv->iterators_tmp = NULL;

	xtv->row_data_func = NULL;
	xtv->row_data_data = NULL;
	xtv->row_params_func = NULL;
	xtv->row_params_data = NULL;
	xtv->data_iterator_func = NULL;
	xtv->data_iterator_data = NULL;

	attach = (gtk_widget_get_direction(GTK_WIDGET(xtv)) == GTK_TEXT_DIR_RTL? 1 : 0);

	gtk_table_resize(GTK_TABLE(xtv), 2, 2);

	xtv->hscrollbar = gtk_hscrollbar_new(NULL);
	gtk_table_attach(GTK_TABLE(xtv), xtv->hscrollbar, attach, attach + 1, 1, 2, GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0 ,0);
	/*gtk_widget_show(xtv->hscrollbar);*/

	xtv->vscrollbar = gtk_vscrollbar_new(NULL);
	gtk_table_attach(GTK_TABLE(xtv), xtv->vscrollbar, 1 - attach, 2 - attach, 0, 1, 
			GTK_SHRINK, GTK_FILL | GTK_EXPAND, 0 ,0);
	gtk_widget_show(xtv->vscrollbar);

	xtv->viewport = gtk_viewport_new(gtk_range_get_adjustment(GTK_RANGE(xtv->hscrollbar)), NULL);
	gtk_table_attach(GTK_TABLE(xtv), xtv->viewport, attach, attach + 1, 0, 1, 
			GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);

	xtv->vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(xtv->viewport), xtv->vbox);

	xtv->hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(xtv->hbox);
	gtk_box_pack_start(GTK_BOX(xtv->vbox), xtv->hbox, FALSE, FALSE, 0);

	xtv->darea = gtk_drawing_area_new();
	gtk_widget_add_events(xtv->darea, GDK_BUTTON_PRESS_MASK |
			GDK_BUTTON_RELEASE_MASK |
			GDK_KEY_PRESS_MASK |
			GDK_KEY_RELEASE_MASK);
	style = gtk_rc_get_style_by_paths(gtk_widget_get_settings(xtv->darea), NULL,
			"*.GtkTreeView", GTK_TYPE_TREE_VIEW);
	gtk_widget_set_style(xtv->darea, style);
	gtk_box_pack_start(GTK_BOX(xtv->vbox), xtv->darea, TRUE, TRUE, 0);
	gtk_widget_show(xtv->darea);

	gtk_widget_show(xtv->viewport);
	gtk_widget_show(xtv->vbox);
#if 0
	g_signal_connect(G_OBJECT(xtv->darea), "size-request",
			 G_CALLBACK(darea_size_request_cb), xtv);
#endif
	g_signal_connect(G_OBJECT(xtv->vbox), "size-allocate",
			 G_CALLBACK(darea_size_allocate_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "configure-event",
			G_CALLBACK(darea_configure_event_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "expose-event",
			G_CALLBACK(darea_expose_event_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->vscrollbar), "value-changed",
			G_CALLBACK(scroll_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "button-press-event",
			G_CALLBACK(darea_button_press_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "button-release-event",
			G_CALLBACK(darea_button_release_cb), xtv);
			
}

static void
darea_size_request_cb(GtkWidget *widget, GtkRequisition *requisition)
{
#ifdef DEBUG
  printf("Size request...\n");
#endif
  requisition->width = DAREA_DEFAULT_WIDTH;
  requisition->height = DAREA_DEFAULT_HEIGHT;
}

static void
darea_size_allocate_cb(GtkWidget *widget, 
		GtkAllocation *allocation, 
		gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	update_resize_grips(xtv);
}

static gboolean
darea_configure_event_cb(GtkWidget *widget, GdkEventConfigure *event, gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
#ifdef DEBUG
	printf("Configure x %d, y %d, w %d, h %d\n", event->x, event->y, event->width, event->height);
#endif

	if (xtv->enabled) {
		darea_fill(xtv, xtv->first_iterator, 0, 0);
	}
	/*gtk_xtree_view_update_adjustment(xtv);*/
		
	return TRUE;
}

static gboolean darea_button_press_cb(GtkWidget *widget, 
		GdkEventButton *event,
		gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	void *selected = NULL;
	void *cache_iterator = NULL;
	gboolean is_selected;
	gboolean is_collapsed;
	gint i = 0;
	
	printf("darea_button_press_cb\n");

	/* button1 = delete old selection and select new row*/
	/* button1 + shift = delete old selection and 
	   select from last focus to new focus */
	/* button1 + ctrl = add to selection */
	
	gtk_xtree_view_find_cell(xtv, (int) event->x, (int) event->y,
			&selected,
			&i, NULL);
	if ((event->state & GDK_CONTROL_MASK) == 0) {
		int j;
		for (j = 0; j <= xtv->last_row; j++)
			xtv->cell_prop[j] &= ~GTK_XTREE_VIEW_SELECTED;
		xtv->total_selected = 0;

	  /*delete_selection(xtv)*/;
	}
	if (event->state & GDK_SHIFT_MASK) {
	} else {
		xtv->focused_iterator = selected;
		cache_iterator = cache_get_row_iterator(xtv, selected);
		cache_get_node_status(xtv, cache_iterator, &is_selected, 
				&is_collapsed);
		cache_set_node_status(xtv, cache_iterator, TRUE, is_collapsed);
		xtv->cell_prop[i] |= GTK_XTREE_VIEW_SELECTED;
		xtv->total_selected = 1;
		gtk_widget_queue_draw(xtv->darea);
	}
	return TRUE;
}

gboolean gtk_xtree_view_find_cell(GtkXTreeView *xtv, 
		gint x, gint y, 
		void **iterator, 
		gint *row, gint *col_id)
{
	gint i;
	gint r;
	gint xmin, xmax;
	GSList *iter;
	GtkXTreeViewColumn *column = NULL;
	
	/* TODO: add a stricter test */
	for (i = 0; i <= xtv->last_row && xtv->y_pos[i] < y; i++) {
	}
	g_return_val_if_fail(i <= xtv->last_row, FALSE);
	
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		xmin = column->widget->allocation.x;
		xmax = xmin + column->widget->allocation.width;
		if (xmin <= x && x < xmax)
			break;
	}
	g_return_val_if_fail(iter, FALSE);
	if (row)
		*row = i;
	if (col_id)
		*col_id = column->col_id;
	if (iterator)
		*iterator = xtv->iterators[i];
	return TRUE;
}

static gboolean darea_button_release_cb(GtkWidget *widget, 
		GdkEventButton *event,
		gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	printf("darea_button_release_cb\n");
	return FALSE;
}

static gboolean darea_expose_event_cb(GtkWidget *widget, GdkEventExpose *event, gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	gint i, col, h, w;
	gint depth;
	gint x, y;
	gint has_children;
	gint is_selected;
	gint is_collapsed;
	gint first_c = 0;
	GSList *first_col = NULL;
	GSList *last_col = NULL;
	GSList *iter;
	gint first_row = -1;
	gint last_row = -1;
	GtkXTreeViewColumn *column;
	void **row_cells;
	GdkRectangle rect;
	GdkWindow *window = xtv->darea->window;
	GdkGC *black_gc = xtv->darea->style->black_gc;
	GdkGC *selected_gc = xtv->darea->style->bg_gc[GTK_STATE_SELECTED];
	GdkGC **fg_gc = xtv->darea->style->fg_gc;
	GtkAllocation *allocation = &xtv->darea->allocation;
	GtkStateType state;
#ifdef DEBUG
	printf("Exposed widget %p, darea %p\n", widget, xtv->darea);
#endif
	
#ifdef DEBUG
	printf("Exposed widget allocation: x = %d, y = %d, width = %d, height = %d\n",
	       allocation->x, allocation->y ,allocation->width, allocation->height);
#endif
#ifdef DEBUG
	printf("Exposing x=%d y=%d w=%d h=%d\n", event->area.x, 
		event->area.y, event->area.width, event->area.height);
#endif
	if (xtv->rows == 0)
		return TRUE;

#ifdef DEBUG 
	if (event->count)
		printf("event->count %d\n", event->count);
#endif
	x = 0;
	y = 0;

	for (col = 0, iter = xtv->columns; iter; iter = iter->next, col++) {
		column = (GtkXTreeViewColumn *) iter->data;
		/*printf("button %d x=%d\n", col, column->widget->allocation.x);*/
		first_col = iter;
		first_c = col;
		if (column->widget->allocation.x + column->widget->allocation.width >= event->area.x)
			
			break;
	}
	for (iter = first_col; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		/*printf(" button %d x=%d\n", col, column->widget->allocation.x);*/
		last_col = iter;
		if (column->widget->allocation.x + column->widget->allocation.width > 
				event->area.x + event->area.width)
			break;
	}
	for (i = 0; i <= xtv->last_row; i++) {
		/*printf("row %d y=%d\n", i, xtv->y_pos[i]);*/
		first_row = i;
		if (xtv->y_pos[i] >= event->area.y)
			break;
	}
	for (i = first_row; i <= xtv->last_row; i++) {
		/*printf(" row %d y=%d\n", i, xtv->y_pos[i]);*/
		last_row = i;
		if (xtv->y_pos[i] > event->area.y + event->area.height)
			break;
	}
	
	for (i = first_row; i <= last_row; i++) {
		/*printf("exposing line: %d, cols: ", i);*/
		is_selected = xtv->cell_prop[i] & GTK_XTREE_VIEW_SELECTED;
		has_children = xtv->cell_prop[i] & GTK_XTREE_VIEW_HAS_CHILDREN;
		is_collapsed = xtv->cell_prop[i] & GTK_XTREE_VIEW_COLLAPSED;
		depth = xtv->cell_prop[i] & GTK_XTREE_VIEW_DEPTH;
		y = (i > 0 ? xtv->y_pos[i - 1] : 0);
		rect.y = y;
		rect.height = xtv->y_pos[i] - y;
		state = (is_selected ? GTK_STATE_SELECTED : GTK_STATE_NORMAL);

		if (is_selected) {
			gdk_draw_rectangle(window,
					selected_gc, 
					TRUE,
					0, y,
					xtv->darea->allocation.width, rect.height);
		}

		row_cells = xtv->cells[i];

		col = first_c;
		for (iter = first_col;/*empty on purpose*/;iter = iter->next) {
			//printf(" %2d", col);
			column = (GtkXTreeViewColumn *) iter->data;
			x = column->widget->allocation.x + 1;
			w = column->widget->allocation.width - 2;
			rect.x = x;
			rect.width = w;
			gdk_gc_set_clip_rectangle(black_gc, &rect);

			if (column->col_id == xtv->expander_col_id) {
				x += depth * xtv->expander_width;
				if (has_children) { 
#if 0
					gdk_draw_pixbuf(window,
							black_gc,
							(is_collapsed ? 
							 xtv->collapsed_pixbuf :
							 xtv->expanded_pixbuf),
							0, 0,
							x, y,
							-1, -1,
							GDK_RGB_DITHER_NORMAL,
							0, 0);
#endif
					gtk_paint_arrow(xtv->darea->style,
							window,
							state,
							GTK_SHADOW_NONE,
							NULL,
							xtv->darea,
							"arrow",
							(is_collapsed ? 
							 GTK_ARROW_RIGHT :
							 GTK_ARROW_DOWN),
							FALSE,
							x, y + (rect.height - xtv->expander_width),
							xtv->expander_width,
							xtv->expander_width);
							
				}
				x += xtv->expander_width + 2;
						
			}
			switch (column->type) {
				case GTK_XCOLUMN_TYPE_TEXT:
					gdk_draw_layout(window, 
							fg_gc[state],
							x, y,
							(PangoLayout *) row_cells[col]);
					break;
				case GTK_XCOLUMN_TYPE_STOCKID:
				case GTK_XCOLUMN_TYPE_PIXBUF:
					if (row_cells[col]) {
					gdk_draw_pixbuf(window,
							black_gc,
							(GdkPixbuf *) row_cells[col],
							0, 0,
							x , y,
							-1, -1,
							GDK_RGB_DITHER_NORMAL, 0, 0);
					}
					break;
				default:
					g_assert_not_reached();
			}
			gdk_gc_set_clip_rectangle(black_gc, NULL);
			if (iter == last_col)
				break;
			col++;
		}
		//printf("\n", i);
		y = xtv->y_pos[i];
	}
	return TRUE;
}

static void darea_fill(GtkXTreeView *xtv, void *first_iterator, guint first_row, guint last_row)
{
	void *iterator;
	void *last_iterator;
	void *cache_iterator;
	gboolean has_children;
	guint depth;
	void *content;
	gchar *attribute;
	gboolean is_collapsed = 0;
	gboolean is_selected = 0;
	GSList *col_iter;
	guint row;
	guint col;
	GtkXTreeViewColumn *column;
	void **row_cells;
	gint height;
	gint width;
	gint h;
	gint w;
	gint x;
	gint y;
	gint debug_i;
	GdkRectangle rect;
	GtkWidget *darea = xtv->darea;
	gboolean not_last;

	/*printf("darea_fill from %d to %d starting with iter %d\n",
			first_row, last_row, (int) first_iterator);*/
#ifdef DEBUG
	printf("darea->allocation = x = %d, y = %d, width = %d, height = %d\n",
	       darea->allocation.x, darea->allocation.y ,darea->allocation.width, darea->allocation.height);
#endif

	if (xtv->length == 0 || xtv->enabled == FALSE) 
		return;

	iterator = first_iterator;
	y = 0;
	row = first_row;
	
	if (row)
		y = xtv->y_pos[row - 1];
	debug_i = 0;
	do {
#ifdef DEBUG
	  printf("row = %d, y = %d\n", row, y);
#endif
		if (row >= xtv->allocated_rows)
			add_row(xtv);

		debug_i++;

		xtv->row_params_func(iterator, 
				&has_children, 
				&depth, 
				xtv->row_params_data);

		cache_iterator = cache_get_row_iterator(xtv, iterator);
		cache_set_row_parameter(xtv, cache_iterator, has_children, depth);
		cache_get_node_status(xtv, cache_iterator, 
				&is_selected,
				&is_collapsed);

		row_cells = xtv->cells[row];
		xtv->iterators[row] = iterator;
		xtv->cell_prop[row] = depth |
			(has_children ? GTK_XTREE_VIEW_HAS_CHILDREN : 0) |
			(is_selected ? GTK_XTREE_VIEW_SELECTED : 0) |
			(is_collapsed ? GTK_XTREE_VIEW_COLLAPSED : 0);

		height = 0;
		for (col = 0, col_iter = xtv->columns; 
				col_iter;
				col_iter = col_iter->next, col++) {
			column = (GtkXTreeViewColumn *) col_iter->data;
			xtv->row_data_func(iterator,
					column->col_id,
					&content,
					&attribute,
					xtv->row_data_data);
			switch (column->type) {
				case GTK_XCOLUMN_TYPE_TEXT:
					pango_layout_set_text
						((PangoLayout *) row_cells[col],
						 (content ? content : ""), -1);
					pango_layout_get_pixel_size
						((PangoLayout *) row_cells[col],
						 NULL, &h);
					if (h > 20)
						printf("layout text: %s %d\n", content, h);
					break;
				case GTK_XCOLUMN_TYPE_PIXBUF:
					row_cells[col] = content;
					if (content) {
						h = gdk_pixbuf_get_height
							((GdkPixbuf *) content);
					} else {
						h = 0;
					}
					break;
				case GTK_XCOLUMN_TYPE_STOCKID:
					if (content) {
						row_cells[col] = 
							gtk_widget_render_icon
							(xtv->darea,
							 (char *) content,
							 GTK_ICON_SIZE_MENU,
							 NULL);
						h = gdk_pixbuf_get_height
							((GdkPixbuf *) row_cells[col]);
						if (h > 20)
							printf("stockid %s %d\n", content, h);
					} else {
						row_cells[col] = NULL;
						h = 0;
					}
					break;
				default:
					g_assert_not_reached();
			}
			height = MAX(height, h);
			if (height > 20)
				printf("height!!! = %d\n", height);
		}
		xtv->y_pos[row] = y + height;
		y += height;
		row++;
#ifdef DEBUG
		printf("row: %d last_row: %d y: %d height: %d\n",
		       row, last_row, y, darea->allocation.height);
#endif
		last_iterator = iterator;
		not_last = xtv->data_iterator_func(&iterator, 
				TRUE, 
				is_collapsed, 
				xtv->sort_col_id, 
				xtv->sort_order, 
				xtv->data_iterator_data);
	} while (row != last_row &&
			y < darea->allocation.height && 
			not_last);
	if (not_last == FALSE || last_row == 0) {
		xtv->last_row = row - 1;
		xtv->last_iterator = last_iterator;
		if (xtv->rows != row) {
			xtv->rows = row;
			/*gtk_xtree_view_update_adjustment(xtv);*/
		}
	}
#ifdef DEBUG
	printf("last_row = %d rows = %d\n", xtv->last_row, xtv->rows);
#endif
}

static void scroll_cb(GtkRange *range, gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	gint increment;
	gboolean is_next;
	gpointer *iterator;
	gint i;
	gint h, w;
	gint y;
	gdouble v = gtk_range_get_value(range);
	gint value = floor(v);
	GdkWindow *window = xtv->darea->window;
	GdkGC *white_gc = xtv->darea->style->white_gc;
	GdkGC *black_gc = xtv->darea->style->black_gc;
	
	gdk_window_freeze_updates(window);
	/*printf("Value %e = %d\n", v, value);*/
	increment = value - xtv->first_value;

	xtv->first_value = value;

	if (increment == 0)
		return;
	
	is_next = (increment > 0);
	if (!is_next)
		increment = -increment;
	
	if (is_next && FALSE) {
		xtv->first_iterator = xtv->iterators[increment];
	} else {
		for (i = 0; i < increment; i++) {
			if (!xtv->data_iterator_func(&xtv->first_iterator,
						is_next,
						FALSE,
						xtv->sort_col_id,
						xtv->sort_order,
						xtv->data_iterator_data))
				break;
		}
	}

	w = xtv->darea->allocation.width;
	h = xtv->darea->allocation.height;
	if (increment >= xtv->rows) {
		darea_fill(xtv, xtv->first_iterator, 0, 0);
		gtk_widget_queue_draw(xtv->darea);
	} else {
		if (is_next) {
			GdkRectangle rect;
			gint j, s, t;
			gint *y_pos = xtv->y_pos;
			void **iterators = xtv->iterators;
			void **row_cells_new;
			void **row_cells_old;
			void *tmp;
			gint last_row = xtv->last_row;
			gint delta = y_pos[increment - 1];
			gint height = y_pos[last_row - 1] - delta;
			gint last_row_height = y_pos[last_row] - 
				y_pos[last_row - 1];

			height = y_pos[last_row - 1] - delta;
 			gdk_draw_drawable(window, 
					black_gc, 
					window,
					0, delta,
					0, 0,
					w, height);

			for (i = 0; i < last_row - increment; i++) {
				y_pos[i] = y_pos[i + increment] - delta;
				iterators[i] = iterators[i + increment];
				xtv->cell_prop[i] =
					xtv->cell_prop[i + increment];
			}
			for (i = 0; i <= last_row - increment; i++) {
				t = (i) % (last_row + 1);
				s = (i + increment) % (last_row + 1);
				row_cells_new = xtv->cells[t];
				row_cells_old = xtv->cells[s];
				for (j = 0; j < xtv->col_nb; j++) {
					tmp = row_cells_new[j];
					row_cells_new[j] = row_cells_old[j];
					row_cells_old[j] = tmp;
				}
			}
			/* Should check children if any */
			darea_fill(xtv, xtv->last_iterator, xtv->last_row - increment, 0);
			/*
			for (i = 0; i <= xtv->last_row; i++) {
				printf("y_pos(%d) = %d\n", i, y_pos[i]);
			}
			printf("rect y=%d h=%d\n", height, h - height);
			 */
			rect.x = 0; rect.y = height; rect.width = w; rect.height = h - height;
			
			gdk_window_invalidate_rect(window, &rect, FALSE);
			
		} else {
			GdkRectangle rect;
			gint i, j, s, t;
			gint height;
			gint delta;
			void **row_cells_new;
			void **row_cells_old;
			void *tmp;

			for (i = 0; i <= xtv->last_row; i++) {
				xtv->y_pos_tmp[i] = xtv->y_pos[i];
				xtv->iterators_tmp[i] = xtv->iterators[i];
				xtv->cell_prop_tmp[i] = xtv->cell_prop[i];
			}
			for (i = xtv->last_row; i >= increment; i--) {
				t = (i - increment) % (xtv->last_row + 1);
				s = (i) % (xtv->last_row + 1);
				row_cells_new = xtv->cells[t];
				row_cells_old = xtv->cells[s];
				for (j = 0; j < xtv->col_nb; j++) {
					tmp = row_cells_new[j];
					row_cells_new[j] = row_cells_old[j];
					row_cells_old[j] = tmp;
				}
			}
			darea_fill(xtv, xtv->first_iterator, 0, increment);
			delta = xtv->y_pos[increment - 1];
			for (i = 0; i <= xtv->last_row; i++) {
				if (i + increment == xtv->allocated_rows)
					add_row(xtv);
				xtv->y_pos[i + increment] = 
					xtv->y_pos_tmp[i] + delta;
				xtv->iterators[i + increment] = 
					xtv->iterators_tmp[i];
				xtv->cell_prop[i + increment] =
					xtv->cell_prop_tmp[i];
				xtv->last_row = i + increment;
				if (xtv->y_pos[i + increment] >= h)
					break;
			}
			xtv->last_iterator = xtv->iterators[xtv->last_row];
			if (xtv->last_row + 1 > xtv->rows) {
#ifdef DEBUG
				printf("scrolling up changed number of displayed rows from %d to %d\n", xtv->rows, xtv->last_row + 1);
#endif
				xtv->rows = xtv->last_row + 1;
			}
			height = h - delta;
			gdk_draw_drawable(window,
					black_gc, 
					window,
					0, 0, 
					0, delta, 
					w, height);
			rect.x = 0; rect.y = 0; rect.width = w; rect.height = delta;
			gdk_window_invalidate_rect(window, &rect, FALSE);
		}
	}
	gdk_window_thaw_updates(window);
	gdk_window_process_updates(window, FALSE);
}
	
static void cache_set_row_parameter(GtkXTreeView *xtv, void *cache_iterator, gboolean has_children, guint depth)
{
	/* Cache these values in the cache*/
}

static void cache_set_node_status(GtkXTreeView *xtv, void *cache_iterator, gboolean is_selected, gboolean is_collapsed)
{
	/* Set these values in the cache */
}

static void cache_get_node_status(GtkXTreeView *xtv, void *cache_iterator, gboolean *is_selected, gboolean *is_collapsed)
{
	/* Get these values from the cache */
	*is_collapsed = FALSE;
	*is_selected = cache_iterator == xtv->focused_iterator;
}

static void *cache_get_row_iterator(GtkXTreeView *xtv, void *iterator)
{
	/* Search in the RB tree for the iterator and return a pointer to the node if found, 
	   else insert a new node and return it */

	return iterator;
}
	
static void add_row(GtkXTreeView *xtv)
{
	GSList *iter;
	void **row_cells;
	GtkXTreeViewColumn *column;
	gint col;
	
	xtv->allocated_rows++;

	xtv->cells = g_renew(void **, xtv->cells, xtv->allocated_rows);
	xtv->cell_prop = g_renew(gint, xtv->cell_prop, xtv->allocated_rows);
	xtv->cell_prop_tmp = g_renew(gint, xtv->cell_prop_tmp, xtv->allocated_rows);
	xtv->y_pos = g_renew(gint, xtv->y_pos, xtv->allocated_rows);
	xtv->y_pos_tmp = g_renew(gint, xtv->y_pos_tmp, xtv->allocated_rows);
	xtv->iterators = g_renew(void *, xtv->iterators, xtv->allocated_rows);
	xtv->iterators_tmp = g_renew(void *, xtv->iterators_tmp, xtv->allocated_rows);

	xtv->cells[xtv->allocated_rows - 1] = row_cells = g_new(void *, g_slist_length(xtv->columns));
	
	for (col = 0, iter = xtv->columns; iter; iter = iter->next, col++) {
		column = (GtkXTreeViewColumn *) iter->data;
		switch(column->type) {
			case GTK_XCOLUMN_TYPE_TEXT:
				row_cells[col] = gtk_widget_create_pango_layout(xtv->darea, NULL);
				break;
			default:
				row_cells[col] = NULL;
		}
	}
}

GtkWidget *gtk_xtree_view_new(void)
{
	return GTK_WIDGET(g_object_new(gtk_xtree_view_get_type(), NULL));
}

gboolean gtk_xtree_view_insert_column(GtkXTreeView *xtv, 
		guint col, 
		guint col_id, 
		GtkXColumnType type, 
		gboolean is_sortable, 
		GtkSortType sort_order, 
		GtkWidget *title_widget, 
		gchar *title_string)
{
	GtkXTreeViewColumn *column;
	gboolean fill;

	g_return_val_if_fail(col <= g_slist_length(xtv->columns), FALSE);
	g_return_val_if_fail(col_id > 0, FALSE);

	column = g_new(GtkXTreeViewColumn, 1);

	column->col_id = col_id;
	column->type = type;
	column->is_sortable = is_sortable;
	column->sort_order = sort_order;
	column->treeview = xtv;
	
	create_title_button(column, title_widget, title_string);
	if (xtv->expansible_col_id == 0 || xtv->expansible_col_id == col_id)
		fill = TRUE;
	else
		fill = FALSE;
	gtk_box_pack_start(GTK_BOX(xtv->hbox), column->widget, fill, TRUE, 0);

	xtv->columns = g_slist_append(xtv->columns, column);
	xtv->col_nb = g_slist_length(xtv->columns);
	return TRUE;
}

static void create_title_button(GtkXTreeViewColumn *column, 
		GtkWidget *title_widget, 
		gchar *title_string)
{
	GtkWidget *button;
	GtkStyle *style;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *arrow;


	button = gtk_button_new();
	style = gtk_rc_get_style_by_paths(gtk_widget_get_settings(button), NULL,
			"*.GtkTreeView.GtkButton", G_OBJECT_TYPE(button));
	gtk_widget_set_style(button, style);

	hbox = gtk_hbox_new(FALSE, 0);
	
	if (title_widget) {
		gtk_box_pack_start(GTK_BOX(hbox), title_widget, FALSE, FALSE, 0);
		gtk_widget_show(title_widget);
	}

	label = gtk_label_new_with_mnemonic(title_string);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);

	arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_IN);
	gtk_box_pack_end(GTK_BOX(hbox), arrow, FALSE, FALSE, 0);

	gtk_container_add(GTK_CONTAINER(button), hbox);
	gtk_widget_show(hbox);

	gtk_widget_show(button);

	column->widget = button;
	column->hbox = hbox;
	column->user_widget = title_widget;
	column->label = label;
	column->arrow = arrow;

	/* Header signals */
	g_signal_connect(G_OBJECT(button), "realize", 
			G_CALLBACK(header_realize_cb), column);
	g_signal_connect(G_OBJECT(column->treeview->hbox), "button-press-event", 
			G_CALLBACK(header_button_press_cb),
			column);
	g_signal_connect(G_OBJECT(button), "button-release-event", 
			G_CALLBACK(header_button_release_cb),
			column);
	g_signal_connect(G_OBJECT(button), "clicked", 
			G_CALLBACK(header_clicked_cb),
			column);
	g_signal_connect(G_OBJECT(button), "event",
			G_CALLBACK(header_event_cb),
			column);
}

static void header_realize_cb(GtkWidget *button, gpointer acolumn)
{
	GdkWindow *resize_grip;
	GdkWindowAttr attr;
	guint attributes_mask;
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;

	attr.window_type = GDK_WINDOW_CHILD;
	attr.wclass = GDK_INPUT_ONLY;
	attr.visual = gtk_widget_get_visual (button);
	attr.colormap = gtk_widget_get_colormap (button);
	attr.event_mask = gtk_widget_get_events (button) |
		(GDK_BUTTON_PRESS_MASK |
		 GDK_BUTTON_RELEASE_MASK |
		 GDK_POINTER_MOTION_MASK |
		 GDK_POINTER_MOTION_HINT_MASK |
		 GDK_KEY_PRESS_MASK);
	attributes_mask = GDK_WA_CURSOR | GDK_WA_X | GDK_WA_Y;
	attr.cursor = gdk_cursor_new_for_display (gtk_widget_get_display (button),
			GDK_SB_H_DOUBLE_ARROW);
	attr.y = 0;
	attr.width = 6;
	attr.height = button->allocation.height;

	attr.x = (button->allocation.x + button->allocation.width) - 3;

	resize_grip = gdk_window_new (button->window,
			&attr, attributes_mask);
	gdk_cursor_unref(attr.cursor);
	gdk_window_show(resize_grip);
	gdk_window_raise(resize_grip);

	column->resize_grip = resize_grip;
	column->being_resized = FALSE;
}

static gboolean header_button_release_cb(GtkWidget *widget, GdkEventButton *event, gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;
	column->being_resized = FALSE;

	return (event->button == 1 ? FALSE : TRUE);
}

static void header_clicked_cb(GtkButton *button, gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;
	GtkXTreeView *xtv = column->treeview;
	GSList *iter;
	GtkSortType sort_order;

	if (xtv->sort_col_id == column->col_id) {
		sort_order = (column->sort_order == GTK_SORT_ASCENDING ?
				GTK_SORT_DESCENDING : GTK_SORT_ASCENDING);
	} else {
		sort_order = column->sort_order;
	}
	gtk_xtree_view_set_sort_column(xtv, column->col_id, sort_order);
	gtk_widget_queue_draw(xtv->darea);
}
	
static gboolean header_button_press_cb(GtkWidget *widget, 
		GdkEventButton *event, gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;
	printf("button press...\n");
	if (event->window == column->resize_grip) {
		printf("in resize window...\n");

		if (gdk_pointer_grab(column->resize_grip, FALSE,
					GDK_POINTER_MOTION_HINT_MASK |
					GDK_BUTTON1_MOTION_MASK |
					GDK_BUTTON_RELEASE_MASK,
					NULL, NULL, event->time))
			return FALSE;
		column->being_resized = TRUE;
		return TRUE;
	}
	return FALSE;
	
}

static gboolean header_event_cb(GtkWidget *widget, 
		GdkEvent *event,
		gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;

	g_return_val_if_fail(event != NULL, FALSE);

	if (event->type == GDK_MOTION_NOTIFY)
		printf(column->being_resized ? "resizin" :"moving...\n");
	return FALSE;
}

void gtk_xtree_view_set_sort_column(GtkXTreeView *xtv, guint col_id, GtkSortType sort_order)
{
	GSList *iter;
	GtkXTreeViewColumn *column;
	
	xtv->sort_col_id = 0;
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		if (column->col_id == col_id) {
			column->sort_order = sort_order;
			xtv->sort_order = sort_order;
			xtv->sort_col_id = col_id;
			gtk_widget_show(column->arrow);
			gtk_arrow_set(GTK_ARROW(column->arrow), 
					(sort_order == GTK_SORT_ASCENDING ?
					 GTK_ARROW_DOWN : 
					 GTK_ARROW_UP),
					GTK_SHADOW_NONE);
		} else {
			gtk_widget_hide(column->arrow);
		}
	}
	g_signal_emit(G_OBJECT(xtv),
			gtkxtreeview_signals[SORT_CHANGED_SIGNAL],
			0);
}
		
void gtk_xtree_view_set_expander_column(GtkXTreeView *xtv, guint col_id)
{
	GSList *iter;
	GtkXTreeViewColumn *column;
	
	xtv->expander_col_id = 0;
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		if (column->col_id == col_id) {
			xtv->expander_col_id = col_id;
		}
	}
}

void gtk_xtree_view_set_column_width(GtkXTreeView *xtv, guint col_id, guint width)
{	
	GSList *iter;
	GtkXTreeViewColumn *column;
	
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		if (column->col_id == col_id) {
			gtk_widget_set_size_request(column->widget, width, -1);
		}
	}

}

void gtk_xtree_view_set_expansible_column(GtkXTreeView *xtv, guint col_id)
{
	GSList *iter;
	GtkXTreeViewColumn *column;
	gboolean fill;

	xtv->expansible_col_id = 0;
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		fill = FALSE;
		if (column->col_id == col_id) {
			xtv->expansible_col_id = col_id;
			fill = TRUE;
		} 
		gtk_box_set_child_packing(GTK_BOX(xtv->hbox),
				column->widget, fill, TRUE, 0, GTK_PACK_START);
		
	}
}

void gtk_xtree_view_set_row_data_func(GtkXTreeView *xtv, GtkXTreeViewRowDataFunc func, gpointer data)
{
	xtv->row_data_func = func;
	xtv->row_data_data = data;
}
void gtk_xtree_view_set_row_params_func(GtkXTreeView *xtv, GtkXTreeViewRowParamsFunc func, gpointer data)
{
	xtv->row_params_func = func;
	xtv->row_params_data = data;
}

void gtk_xtree_view_set_data_iterator_func(GtkXTreeView *xtv, GtkXTreeViewDataIteratorFunc func, gpointer data)
{
	xtv->data_iterator_func = func;
	xtv->data_iterator_data = data;
}


void gtk_xtree_view_set_first_row(GtkXTreeView *xtv, void *iterator)
{
	xtv->first_iterator = iterator;
	gtk_range_set_value(GTK_RANGE(xtv->vscrollbar), 0.0);
}

void gtk_xtree_view_freeze(GtkXTreeView *xtv)
{
	xtv->enabled = FALSE;
}

void gtk_xtree_view_thaw(GtkXTreeView *xtv)
{
	xtv->enabled = TRUE;
	gtk_xtree_view_update(xtv);
}


void gtk_xtree_view_set_length(GtkXTreeView *xtv, guint length)
{
	xtv->length = length;
	gtk_xtree_view_update_adjustment(xtv);
}

static void gtk_xtree_view_update_adjustment(GtkXTreeView *xtv)
{
	GtkAdjustment *adj;
	
	adj = gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));

	adj->lower = 0.0;
	adj->upper = xtv->length;
	adj->step_increment = 1.0;
	adj->page_increment = xtv->length/20;/*(xtv->rows > 1 ? xtv->rows - 1 : 1.0);*/
	adj->page_size = 1; /*(xtv->rows > 1 ? xtv->rows - 1 : 1.0);*/
	gtk_adjustment_changed(adj);
	adj = gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));
}

void gtk_xtree_expand_all(GtkXTreeView *xtv)
{
}

static void update_resize_grips(GtkXTreeView *xtv)
{
	GSList *iter;
	GtkXTreeViewColumn *column;

	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		if (GTK_WIDGET_REALIZED(column->widget)) {
			gdk_window_move(column->resize_grip,
					column->widget->allocation.x +
					column->widget->allocation.width - 3,
					0);
			gdk_window_show(column->resize_grip);
			gdk_window_raise(column->resize_grip);
		}
	}
}

void gtk_xtree_view_update(GtkXTreeView *xtv)
{
	darea_fill(xtv, xtv->first_iterator, 0, 0);
	gtk_widget_queue_draw(xtv->darea);
}

void *gtk_xtree_view_get_focused(GtkXTreeView *xtv)
{
	return xtv->focused_iterator;
}

uint32_t gtk_xtree_view_get_selection(GtkXTreeView *xtv, 
		void ***selection)
{
	if (selection) {
		void *iterator = xtv->first_iterator;
		void *cache_iterator;
		void **sel = NULL;
		uint32_t i = 0;
		gboolean is_selected;
		gboolean is_collapsed;
		
		sel = g_new(void, xtv->total_selected);
		
		do {
			cache_iterator = cache_get_row_iterator(xtv, iterator);
			cache_get_node_status(xtv, cache_iterator, &is_selected,
					&is_collapsed);
			if (is_selected) {
				sel[i] = iterator;
				i++;
			}
		} while (i < xtv->total_selected &&
				xtv->data_iterator_func(&iterator,
					TRUE,
					FALSE,
					xtv->sort_col_id,
					xtv->sort_order,
					xtv->data_iterator_data));

		if (i < xtv->total_selected) {
			uint32_t j;
			for (j = 0; j < i; j++) {
				sel[xtv->total_selected - 1 -j] = sel[j];
			}
			j = 0;
			iterator = xtv->focused_iterator;
			while (xtv->data_iterator_func(&iterator, FALSE, FALSE,
						xtv->sort_col_id,
						xtv->sort_order,
						xtv->data_iterator_data)
					&& j < i) {
				cache_iterator = cache_get_row_iterator(xtv,
						iterator);
				cache_get_node_status(xtv,
						cache_iterator,
						&is_selected,
						&is_collapsed);
				if (is_selected) {
					sel[j] = iterator;
					j++;
				}
			}
			if (j != i) {
				printf("could not found the correct number of selected items\n");
			}
		}
		*selection = sel;
	}

	return xtv->total_selected;
}
