/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This is part of XetPan
 *
 * This is a tree view/list view widget taylored for small memory footprint
 * and for direct accessing of user data via user provided function
 * to retrieve data and to navigate through it.
 * 
 * Some parts of this code are heavily inspired from gtk+ 2.x gtktreeview,
 * for example: resize grips in header button, some function names.
 *
 * Gtk+ is to be found at http://www.gtk.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef GTKXTREEVIEW_H
#define GTKXTREEVIEW_H

#include <inttypes.h>
#include <glib.h>
#include <glib-object.h>

G_BEGIN_DECLS

#define GTK_XTREE_VIEW_TYPE		(gtk_xtree_view_get_type ())
#define GTK_XTREE_VIEW(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_XTREE_VIEW_TYPE, GtkXTreeView))
#define GTK_XTREE_VIEW_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), GTK_XTREE_VIEW_TYPE, GtkXTreeViewClass))
#define IS_GTK_XTREE_VIEW(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTK_XTREE_VIEW_TYPE))
#define IS_GTK_XTREE_VIEW_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GTK_XTREE_VIEW_TYPE))

typedef struct _GtkXTreeView      GtkXTreeView;
typedef struct _GtkXTreeViewClass GtkXTreeViewClass;

typedef enum
{
	GTK_XCOLUMN_TYPE_TEXT,
	GTK_XCOLUMN_TYPE_PIXBUF,
	GTK_XCOLUMN_TYPE_STOCKID
} GtkXColumnType;

/*! The function that returns the data associated with a row
 * \param iterator a pointer to the data
 * \param col the column ID for which data is needed.
 * \param text a place holder which will contain the pointer to the data
 * \param attribute a place holder which will contain the name of the attribute
 *        that should be used to display the data
 */
typedef gboolean (*GtkXTreeViewRowDataFunc)(void *iterator, guint col_id, void **content, gchar **attribute, gpointer data);
	 	
/*! Get a params of a node: whether it has children, and its depth in the tree
 */
  
typedef void (*GtkXTreeViewRowParamsFunc)(void *iterator, gboolean *has_children, guint *depth, gpointer data);

/*! A function to traverse the data, step by step on a certain direction
 *  \param iterator the starting point
 *  \param whether we want the next iterator or the previous one
 *  \param ignore_children whether we do not iterate through children
 *  \param sort_column_id the id of the sorting column
 *  \param sort_order the sort order
 *  \return TRUE if next iterator was found and iterator is changed; FALSE if
 *          no other iterator was found, and iterator is not changed
 *
 *  The idea is that the user can better sort its data and optionnally cache sorting results
 *  When the widget issues a "sort-changed" signal, the user must catch it and sort its data
 *  so that when the next time this function is called it is ready to return the requested 
 *  iterator
 */
typedef gboolean (*GtkXTreeViewDataIteratorFunc)(void **iterator, gboolean next, gboolean ignore_children, guint sort_col_id, GtkSortType sort_order, gpointer data);

GType gtk_xtree_view_get_type(void);

GtkWidget *gtk_xtree_view_new(void);

/*! Insert a new column
 *  \param col column number where the new column will be inserted. 0 to prepend.
 *  \param type the type of the column GTK_XCOLUMN_TYPE_TEXT, GTK_XCOLUMN_TYPE_PIXBUF
 *  \param column_id a user provided id to identify the column without its number
 *  \param title_widget a widget that will put before the column title_string
 *  \param title_string the title of the column
 */
gboolean gtk_xtree_view_insert_column(GtkXTreeView *xtv, guint col, guint col_id, GtkXColumnType type, gboolean is_sortable, GtkSortType sort_order, GtkWidget *title_widget, gchar *title_string);

void gtk_xtree_view_set_sort_column(GtkXTreeView *xtv, guint col_id, GtkSortType sort_order);

void gtk_xtree_view_set_column_width(GtkXTreeView *xtv, guint col_id, guint width);

void gtk_xtree_view_set_expansible_column(GtkXTreeView *xtv, guint col_id);

void gtk_xtree_view_set_expander_column(GtkXTreeView *xtv, guint col_id);

/*! Set the function that will return the row content
 *  \param func the function
 *  \param data a user data passed to func
 */
void gtk_xtree_view_set_row_data_func(GtkXTreeView *xtv, GtkXTreeViewRowDataFunc func, gpointer data);

void gtk_xtree_view_set_row_params_func(GtkXTreeView *xtv, GtkXTreeViewRowParamsFunc func, gpointer data);

void gtk_xtree_view_set_data_iterator_func(GtkXTreeView *xtv, GtkXTreeViewDataIteratorFunc func, gpointer data);

void gtk_xtree_view_update_all(GtkXTreeView *xtv);

void gtk_xtree_view_update_row(GtkXTreeView *xtv, void *iterator);

void gtk_xtree_view_update_cell(GtkXTreeView *xtv, void *iterator, guint col_id);

void gtk_xtree_view_set_first_row(GtkXTreeView *xtv, void *iterator);
void gtk_xtree_view_set_length(GtkXTreeView *xtv, guint length);

void gtk_xtree_view_freeze(GtkXTreeView *xtv);
void gtk_xtree_view_thaw(GtkXTreeView *xtv);
void gtk_xtree_view_update(GtkXTreeView *xtv);
void *gtk_xtree_view_get_focused(GtkXTreeView *xtv);
gboolean gtk_xtree_view_find_cell(GtkXTreeView *xtv, 
		gint x, gint y, 
		void **iterator,
		gint *row, 
		gint *col_id);
uint32_t gtk_xtree_view_get_selection(GtkXTreeView *xtv, 
		void ***selection);
		
G_END_DECLS

#endif
