/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>  /* open */
#include <sys/stat.h>   /* open, stat */
#include <fcntl.h>      /* open */

#include <unistd.h>     /* dup, stat */
#include <errno.h>

#include "xetpan.h"
#include "logging.h"
#include "conf.h"
#include "ui.h"

static void
Logging_write(int fd, short int condition, void *data);

int
a_Logging_redirect(XetPan *xetpan, const char *path)
{
  struct stat buf;
  char *logfile;
  char *logbackup;
  int log_fd;
  int ret;
  int pipe_fd[2];

  logfile   = a_Conf_create_filename(path, XETPAN_LOG_FILE);
  logbackup = a_Conf_create_filename(path, XETPAN_LOG_FILE_BAK);

  ret = stat(logfile, &buf);
  if (ret == 0)
    {
      if (rename(logfile, logbackup))
	{
	  A_LOGGING_ERROR("rename(%s, %s): %s", logfile, logbackup, 
			  strerror(errno));
	  goto err;
	}
    }
  else if (errno != ENOENT)
    {
      A_LOGGING_ERROR("stat(%s): %s", logfile, strerror(errno));
      goto err;
    }
  
  log_fd = open(logfile, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
  if (log_fd == -1)
    {
      A_LOGGING_ERROR("open(%s): %s", logfile, strerror(errno));
      goto err;
    }

  xetpan->log_fd = log_fd;

  A_LOGGING_INFO("Logging redirected to file '%s'", logfile);

  if (pipe(pipe_fd))
    {
      A_LOGGING_SYSTEM_ERROR("pipe: %s", strerror(errno));
      goto err;
    }

  close(fileno(stderr));
  dup(pipe_fd[1]);
  a_UI_add_io_watch(xetpan, pipe_fd[0],
		    XETPAN_POLLIN | 
		    XETPAN_POLLHUP |
		    XETPAN_POLLERR,
		    Logging_write,
		    xetpan);
  free(logfile);
  free(logbackup);

  return 0;

err:
  A_LOGGING_ERROR("logging will not be redirected");
  free(logfile);
  free(logbackup);
  return errno;
}

static void
Logging_write(int fd, short int condition, void *data)
{
  XetPan *xetpan = (XetPan *) data;
  int log_fd = xetpan->log_fd;
  int notify_ui = 0;

  if (condition & (XETPAN_POLLHUP | XETPAN_POLLERR))
    {
      const char *err_msg = "POLLHUP | POLLERR on stderr...\n";
      write(log_fd, err_msg, strlen(err_msg));
      notify_ui = 1;
    }
  else
    {
      char buf[8192];
      int count;
      count = read(fd, buf, sizeof(buf));
      if (count > 0)
	{
	  int wcount;
	  wcount = write(log_fd, buf, count);
	  a_UI_notify_error(xetpan);
	}
    }
}
